import pandas as pd

from virny.configs.constants import ModelSetting
from virny.custom_classes.base_dataset import BaseFlowDataset
from virny.analyzers.subgroup_variance_calculator import SubgroupVarianceCalculator
from virny.analyzers.batch_overall_variance_analyzer import BatchOverallVarianceAnalyzer
from virny.analyzers.incremental_overall_variance_analyzer import IncrementalOverallVarianceAnalyzer


class SubgroupVarianceAnalyzer:
    """
    Analyzer to compute variance metrics for subgroups.

    Parameters
    ----------
    model_setting
        Model learning type; a constant from virny.configs.constants.ModelSetting
    n_estimators
        Number of estimators for bootstrap
    base_model
        Initialized base model to analyze
    base_model_name
        Model name
    bootstrap_fraction
        [0-1], fraction from train_pd_dataset for fitting an ensemble of base models
    dataset
        Initialized object of GenericPipeline class
    dataset_name
        Name of dataset, used for correct results naming
    sensitive_attributes_dct
        A dictionary where keys are sensitive attribute names (including attributes intersections),
         and values are privilege values for these attributes
    test_protected_groups
        A dictionary of protected groups where keys are subgroup names,
         and values are X_test row indexes correspondent to this subgroup.
    computation_mode
        [Optional] A non-default mode for metrics computation. Should be included in the ComputationMode enum.
    verbose
        [Optional] Level of logs printing. The greater level provides more logs.
         As for now, 0, 1, 2 levels are supported.

    """
    def __init__(self, model_setting: ModelSetting, n_estimators: int, base_model, base_model_name: str,
                 bootstrap_fraction: float, dataset: BaseFlowDataset, dataset_name: str,
                 sensitive_attributes_dct: dict, test_protected_groups: dict, computation_mode: str = None, verbose: int = 0):
        if model_setting == ModelSetting.BATCH:
            overall_variance_analyzer = BatchOverallVarianceAnalyzer(base_model=base_model,
                                                                     base_model_name=base_model_name,
                                                                     bootstrap_fraction=bootstrap_fraction,
                                                                     X_train=dataset.X_train_val,
                                                                     y_train=dataset.y_train_val,
                                                                     X_test=dataset.X_test,
                                                                     y_test=dataset.y_test,
                                                                     dataset_name=dataset_name,
                                                                     target_column=dataset.target,
                                                                     n_estimators=n_estimators,
                                                                     verbose=verbose)
        elif model_setting == ModelSetting.INCREMENTAL:
            overall_variance_analyzer = IncrementalOverallVarianceAnalyzer(base_model=base_model,
                                                                           base_model_name=base_model_name,
                                                                           bootstrap_fraction=bootstrap_fraction,
                                                                           X_train=dataset.X_train_val,
                                                                           y_train=dataset.y_train_val,
                                                                           X_test=dataset.X_test,
                                                                           y_test=dataset.y_test,
                                                                           dataset_name=dataset_name,
                                                                           target_column=dataset.target,
                                                                           n_estimators=n_estimators,
                                                                           verbose=verbose)
        else:
            raise ValueError('model_setting is incorrect or not supported')

        self.dataset_name = overall_variance_analyzer.dataset_name
        self.n_estimators = overall_variance_analyzer.n_estimators
        self.base_model_name = overall_variance_analyzer.base_model_name

        self.__overall_variance_analyzer = overall_variance_analyzer
        self.__subgroup_variance_calculator = SubgroupVarianceCalculator(X_test=dataset.X_test,
                                                                         y_test=dataset.y_test,
                                                                         sensitive_attributes_dct=sensitive_attributes_dct,
                                                                         test_protected_groups=test_protected_groups,
                                                                         computation_mode=computation_mode)
        self.overall_variance_metrics_dct = dict()
        self.subgroup_variance_metrics_dct = dict()

    def set_test_sets(self, new_X_test, new_y_test):
        self.__overall_variance_analyzer.X_test = new_X_test
        self.__overall_variance_analyzer.y_test = new_y_test
        self.__subgroup_variance_calculator.X_test = new_X_test
        self.__subgroup_variance_calculator.y_test = new_y_test

    def set_test_protected_groups(self, new_test_protected_groups):
        self.__subgroup_variance_calculator.test_protected_groups = new_test_protected_groups

    def compute_metrics(self, save_results: bool, result_filename: str = None, save_dir_path: str = None,
                        make_plots: bool = True, with_fit: bool = True):
        """
        Measure variance metrics for subgroups for the base model. Display variance plots for analysis if needed.
         Save results to a .csv file if needed.

        Return averaged bootstrap predictions and a pandas dataframe of variance metrics for subgroups.

        Parameters
        ----------
        save_results
            If we need to save result metrics in a file
        result_filename
            [Optional] Filename for results to save
        save_dir_path
            [Optional] Location where to save the results file
        make_plots
            If to display plots for analysis
        with_fit
            If to fit estimators in bootstrap

        """
        y_preds, y_test_true = self.__overall_variance_analyzer.compute_metrics(make_plots, save_results=False, with_fit=with_fit)
        self.overall_variance_metrics_dct = self.__overall_variance_analyzer.get_metrics_dict()

        # Count and display fairness metrics
        self.__subgroup_variance_calculator.set_overall_variance_metrics(self.overall_variance_metrics_dct)
        self.subgroup_variance_metrics_dct = self.__subgroup_variance_calculator.compute_subgroup_metrics(
            self.__overall_variance_analyzer.models_predictions, save_results, result_filename, save_dir_path
        )

        return y_preds, pd.DataFrame(self.subgroup_variance_metrics_dct)



import pathlib
import pandas as pd
import numpy as np

from sklearn import preprocessing
from folktables import ACSDataSource, ACSEmployment, ACSIncome, ACSTravelTime, ACSPublicCoverage, ACSMobility, \
    employment_filter, adult_filter, public_coverage_filter
from virny.datasets.base import BaseDataLoader


class CreditDataset(BaseDataLoader):
    """
    Dataset class for the Credit dataset that contains sensitive attributes among feature columns.
    Source: https://www.kaggle.com/competitions/GiveMeSomeCredit/overview

    Parameters
    ----------
    subsample_size
        Subsample size to create based on the input dataset
    subsample_seed
        Seed for sampling using the sample() method from pandas

    """
    def __init__(self, subsample_size: int = None, subsample_seed: int = None):
        filename = 'givemesomecredit.csv'
        dataset_path = pathlib.Path(__file__).parent.joinpath(filename)

        df = pd.read_csv(dataset_path, index_col=0)
        if subsample_size:
            df = df.sample(subsample_size, random_state=subsample_seed) if subsample_seed is not None \
                else df.sample(subsample_size)
            df = df.reset_index(drop=True)

        target = 'SeriousDlqin2yrs'
        numerical_columns = ['RevolvingUtilizationOfUnsecuredLines', 'age', 'NumberOfTime30-59DaysPastDueNotWorse',
                             'DebtRatio', 'MonthlyIncome', 'NumberOfOpenCreditLinesAndLoans',
                             'NumberOfTimes90DaysLate', 'NumberRealEstateLoansOrLines']
        categorical_columns = []

        super().__init__(
            full_df=df,
            target=target,
            numerical_columns=numerical_columns,
            categorical_columns=categorical_columns,
        )


class LawSchoolDataset(BaseDataLoader):
    """
    Dataset class for the Law School dataset that contains sensitive attributes among feature columns.
    Source: https://github.com/tailequy/fairness_dataset/blob/main/experiments/data/law_school_clean.csv
    Description: https://arxiv.org/pdf/2110.00530.pdf

    Parameters
    ----------
    subsample_size
        Subsample size to create based on the input dataset
    subsample_seed
        Seed for sampling using the sample() method from pandas
    dataset_path
        [Optional] Path to a file with the data

    """
    def __init__(self, subsample_size: int = None, subsample_seed: int = None, dataset_path=None):
        if dataset_path is None:
            filename = 'law_school_clean.csv'
            dataset_path = pathlib.Path(__file__).parent.joinpath(filename)

        df = pd.read_csv(dataset_path)
        if subsample_size:
            df = df.sample(subsample_size, random_state=subsample_seed) if subsample_seed is not None \
                else df.sample(subsample_size)
            df = df.reset_index(drop=True)

        # Cast columns
        columns_to_cast = ['fulltime', 'fam_inc', 'male', 'tier']
        columns_to_cast_dct = {col: "str" for col in columns_to_cast}
        df = df.astype(columns_to_cast_dct)

        target = 'pass_bar'
        numerical_columns = ['decile1b', 'decile3', 'lsat', 'ugpa', 'zfygpa', 'zgpa']
        categorical_columns = ['fulltime', 'fam_inc', 'male', 'tier', 'race']

        super().__init__(
            full_df=df,
            target=target,
            numerical_columns=numerical_columns,
            categorical_columns=categorical_columns,
        )


class DiabetesDataset(BaseDataLoader):
    """
    Dataset class for the Diabetes dataset that contains sensitive attributes among feature columns.
    Source: https://github.com/tailequy/fairness_dataset/blob/main/experiments/data/diabetes-clean.csv
    Description: https://arxiv.org/pdf/2110.00530.pdf

    Parameters
    ----------
    subsample_size
        Subsample size to create based on the input dataset
    subsample_seed
        Seed for sampling using the sample() method from pandas
    dataset_path
        [Optional] Path to a file with the data

    """
    def __init__(self, subsample_size: int = None, subsample_seed: int = None, dataset_path=None):
        if dataset_path is None:
            filename = 'diabetes-clean.csv'
            dataset_path = pathlib.Path(__file__).parent.joinpath(filename)

        df = pd.read_csv(dataset_path)
        if subsample_size:
            df = df.sample(subsample_size, random_state=subsample_seed) if subsample_seed is not None \
                else df.sample(subsample_size)
            df = df.reset_index(drop=True)

        # Cast columns
        columns_to_cast = ['admission_type_id', 'discharge_disposition_id', 'admission_source_id']
        columns_to_cast_dct = {col: "str" for col in columns_to_cast}
        df = df.astype(columns_to_cast_dct)
        df = df.drop(['encounter_id', 'patient_nbr'], axis=1)

        # Encode labels
        le = preprocessing.LabelEncoder()
        for i in ['diag_1', 'diag_2', 'diag_3']:
            df[i] = le.fit_transform(df[i])

        target = 'readmitted'
        df[target] = df[target].replace(['<30'], 1)
        df[target] = df[target].replace(['>30'], 0)
        numerical_columns = ['number_diagnoses', 'time_in_hospital', 'num_lab_procedures', 'num_procedures',
                             'num_medications', 'number_outpatient', 'number_emergency', 'number_inpatient']
        categorical_columns = [
            'race',	'gender', 'age', 'admission_type_id', 'discharge_disposition_id', 'admission_source_id',
            'diag_1', 'diag_2', 'diag_3', 'max_glu_serum', 'A1Cresult', 'metformin', 'repaglinide', 'nateglinide',
            'chlorpropamide', 'glimepiride', 'acetohexamide', 'glipizide', 'glyburide', 'tolbutamide', 'pioglitazone',
            'rosiglitazone', 'acarbose', 'miglitol',	'troglitazone', 'tolazamide', 'examide', 'citoglipton',
            'insulin', 'glyburide-metformin', 'glipizide-metformin', 'glimepiride-pioglitazone',
            'metformin-rosiglitazone', 'metformin-pioglitazone', 'change', 'diabetesMed'
        ]

        super().__init__(
            full_df=df,
            target=target,
            numerical_columns=numerical_columns,
            categorical_columns=categorical_columns,
        )


class RicciDataset(BaseDataLoader):
    """
    Dataset class for the Ricci dataset that contains sensitive attributes among feature columns.
    Source: https://github.com/tailequy/fairness_dataset/blob/main/experiments/data/ricci_race.csv
    Description: https://arxiv.org/pdf/2110.00530.pdf

    Parameters
    ----------
    dataset_path
        [Optional] Path to a file with the data

    """
    def __init__(self, dataset_path=None):
        if dataset_path is None:
            filename = 'ricci_race.csv'
            dataset_path = pathlib.Path(__file__).parent.joinpath(filename)

        df = pd.read_csv(dataset_path)

        target = 'Promoted'
        df[target] = df[target].replace([-1], 0)
        numerical_columns = ['Oral', 'Written', 'Combine']
        categorical_columns = ['Position', 'Race']

        super().__init__(
            full_df=df,
            target=target,
            numerical_columns=numerical_columns,
            categorical_columns=categorical_columns,
        )


class CompasDataset(BaseDataLoader):
    """
    Dataset class for the COMPAS dataset that contains sensitive attributes among feature columns.

    Parameters
    ----------
    subsample_size
        Subsample size to create based on the input dataset
    subsample_seed
        Seed for sampling using the sample() method from pandas
    dataset_path
        [Optional] Path to a file with the data

    """
    def __init__(self, subsample_size: int = None, subsample_seed: int = None, dataset_path=None):
        if dataset_path is None:
            filename = 'COMPAS.csv'
            dataset_path = pathlib.Path(__file__).parent.joinpath(filename)

        df = pd.read_csv(dataset_path)
        if subsample_size:
            df = df.sample(subsample_size, random_state=subsample_seed) if subsample_seed is not None \
                else df.sample(subsample_size)
            df = df.reset_index(drop=True)

        int_columns = ['recidivism', 'age', 'age_cat_25 - 45', 'age_cat_Greater than 45',
                       'age_cat_Less than 25', 'c_charge_degree_F', 'c_charge_degree_M', 'sex']
        int_columns_dct = {col: "int" for col in int_columns}
        df = df.astype(int_columns_dct)

        target = 'recidivism'
        numerical_columns = ['age', 'juv_fel_count', 'juv_misd_count', 'juv_other_count', 'priors_count']
        categorical_columns = ['race', 'age_cat_25 - 45', 'age_cat_Greater than 45',
                               'age_cat_Less than 25', 'c_charge_degree_F', 'c_charge_degree_M', 'sex']

        super().__init__(
            full_df=df,
            target=target,
            numerical_columns=numerical_columns,
            categorical_columns=categorical_columns,
        )


class CompasWithoutSensitiveAttrsDataset(BaseDataLoader):
    """
    Dataset class for the COMPAS dataset that does not contain sensitive attributes among feature columns
     to test blind classifiers

    Parameters
    ----------
    subsample_size
        Subsample size to create based on the input dataset
    subsample_seed
        Seed for sampling using the sample() method from pandas
    dataset_path
        [Optional] Path to a file with the data

    """
    def __init__(self, subsample_size: int = None, subsample_seed: int = None, dataset_path=None):
        if dataset_path is None:
            filename = 'COMPAS.csv'
            dataset_path = pathlib.Path(__file__).parent.joinpath(filename)

        df = pd.read_csv(dataset_path)
        if subsample_size:
            df = df.sample(subsample_size, random_state=subsample_seed) if subsample_seed is not None \
                else df.sample(subsample_size)
            df = df.reset_index(drop=True)

        # Initial data types transformation
        int_columns = ['recidivism', 'age', 'age_cat_25 - 45', 'age_cat_Greater than 45',
                       'age_cat_Less than 25', 'c_charge_degree_F', 'c_charge_degree_M', 'sex']
        int_columns_dct = {col: "int" for col in int_columns}
        df = df.astype(int_columns_dct)

        # Define params
        target = 'recidivism'
        numerical_columns = ['juv_fel_count', 'juv_misd_count', 'juv_other_count','priors_count']
        categorical_columns = ['age_cat_25 - 45', 'age_cat_Greater than 45','age_cat_Less than 25',
                               'c_charge_degree_F', 'c_charge_degree_M']

        super().__init__(
            full_df=df,
            target=target,
            numerical_columns=numerical_columns,
            categorical_columns=categorical_columns
        )


class ACSIncomeDataset(BaseDataLoader):
    """
    Dataset class for the income task from the folktables dataset.
    Target: binary classification, predict if a person has an annual income > $50,000.
    Source of the dataset: https://github.com/socialfoundations/folktables

    Parameters
    ----------
    state
        State in the US for which to get the data. All states in the US are available.
    year
        Year for which to get the data. Five different years of data collection are available: 2014–2018 inclusive.
    root_dir
        Path to the root directory where to store the extracted dataset or where it is stored.
    with_nulls
        Whether to keep nulls in the dataset or replace them on the new categorical class. Default: False.
    with_filter
        Whether to use a folktables filter for this task. Default: True.
    optimize
        Whether to optimize the dataset size by downcasting categorical columns. Default: True.
    subsample_size
        Subsample size to create based on the input dataset.
    subsample_seed
        Seed for sampling using the sample() method from pandas.

    """
    def __init__(self, state, year, root_dir=None, with_nulls=False, with_filter=True,
                 optimize=True, subsample_size: int = None, subsample_seed: int = None):
        data_dir = pathlib.Path(__file__).parent if root_dir is None else root_dir
        data_source = ACSDataSource(
            survey_year=year,
            horizon='1-Year',
            survey='person',
            root_dir=data_dir
        )
        acs_data = data_source.get_data(states=state, download=True)
        if with_filter:
            acs_data = adult_filter(acs_data)
        if subsample_size:
            acs_data = acs_data.sample(subsample_size, random_state=subsample_seed) if subsample_seed is not None \
                else acs_data.sample(subsample_size)
            acs_data = acs_data.reset_index(drop=True)

        features = ACSIncome.features
        target = ACSIncome.target
        categorical_columns = ['SCHL', 'COW', 'MAR', 'OCCP', 'POBP', 'RELP', 'SEX', 'RAC1P']
        numerical_columns = ['AGEP', 'WKHP']

        if with_nulls:
            X_data = acs_data[features]
        else:
            X_data = acs_data[features].apply(lambda x: np.nan_to_num(x, -1))

        if optimize:
            X_data = optimize_data_loading(X_data, categorical_columns)

        optimized_X_data = X_data[categorical_columns].astype('str')
        for col in numerical_columns:
            optimized_X_data[col] = X_data[col]
        y_data = acs_data[target].apply(lambda x: int(x > 50_000))

        columns_with_nulls = optimized_X_data.columns[optimized_X_data.isna().any().to_list()].to_list()

        super().__init__(
            full_df=optimized_X_data,
            target=target,
            numerical_columns=numerical_columns,
            categorical_columns=categorical_columns,
            X_data=optimized_X_data,
            y_data=y_data,
            columns_with_nulls=columns_with_nulls,
        )

    def update_X_data(self, X_data):
        """
        To save simulated nulls
        """
        self.X_data = X_data
        self.columns_with_nulls = self.X_data.columns[self.X_data.isna().any().to_list()].to_list()


class ACSEmploymentDataset(BaseDataLoader):
    """
    Dataset class for the employment task from the folktables dataset.
    Target: binary classification, predict if a person is employed.
    Source of the dataset: https://github.com/socialfoundations/folktables

    Parameters
    ----------
    state
        State in the US for which to get the data. All states in the US are available.
    year
        Year for which to get the data. Five different years of data collection are available: 2014–2018 inclusive.
    root_dir
        Path to the root directory where to store the extracted dataset or where it is stored.
    with_nulls
        Whether to keep nulls in the dataset or replace them on the new categorical class. Default: False.
    with_filter
        Whether to use a folktables filter for this task. Default: True.
    optimize
        Whether to optimize the dataset size by downcasting categorical columns. Default: True.
    subsample_size
        Subsample size to create based on the input dataset.
    subsample_seed
        Seed for sampling using the sample() method from pandas.

    """
    def __init__(self, state, year, root_dir=None, with_nulls=False, with_filter=True,
                 optimize=True, subsample_size: int = None, subsample_seed: int = None):
        data_dir = pathlib.Path(__file__).parent if root_dir is None else root_dir
        data_source = ACSDataSource(
            survey_year=year,
            horizon='1-Year',
            survey='person',
            root_dir=data_dir
        )
        acs_data = data_source.get_data(states=state, download=True)
        if with_filter:
            acs_data = employment_filter(acs_data)
        if subsample_size:
            acs_data = acs_data.sample(subsample_size, random_state=subsample_seed) if subsample_seed is not None \
                else acs_data.sample(subsample_size)
            acs_data = acs_data.reset_index(drop=True)

        features = ACSEmployment.features
        target = ACSEmployment.target
        categorical_columns = ['MAR', 'MIL', 'ESP', 'MIG', 'DREM', 'NATIVITY', 'DIS', 'DEAR', 'DEYE', 'SEX', 'RAC1P', 'RELP', 'CIT', 'ANC','SCHL']
        numerical_columns = ['AGEP']

        if with_nulls is True:
            X_data = acs_data[features]
        else:
            X_data = acs_data[features].apply(lambda x: np.nan_to_num(x, -1))

        if optimize:
            X_data = optimize_data_loading(X_data, categorical_columns)

        optimized_X_data = X_data[categorical_columns].astype('str')
        for col in numerical_columns:
            optimized_X_data[col] = X_data[col]
        y_data = acs_data[target].apply(lambda x: int(x == 1))

        columns_with_nulls = optimized_X_data.columns[optimized_X_data.isna().any().to_list()].to_list()

        super().__init__(
            full_df=optimized_X_data,
            target=target,
            numerical_columns=numerical_columns,
            categorical_columns=categorical_columns,
            X_data=optimized_X_data,
            y_data=y_data,
            columns_with_nulls=columns_with_nulls,
        )

    def update_X_data(self, X_data):
        """
        To save simulated nulls
        """
        self.X_data = X_data
        self.columns_with_nulls = self.X_data.columns[self.X_data.isna().any().to_list()].to_list()


class ACSMobilityDataset(BaseDataLoader):
    """
    Dataset class for the mobility task from the folktables dataset.
    Target: binary classification, predict whether a young adult moved addresses in the last year.
    Source of the dataset: https://github.com/socialfoundations/folktables

    Parameters
    ----------
    state
        State in the US for which to get the data. All states in the US are available.
    year
        Year for which to get the data. Five different years of data collection are available: 2014–2018 inclusive.
    root_dir
        Path to the root directory where to store the extracted dataset or where it is stored.
    with_nulls
        Whether to keep nulls in the dataset or replace them on the new categorical class. Default: False.

    """
    def __init__(self, state, year, root_dir=None, with_nulls=False):
        data_dir = pathlib.Path(__file__).parent if root_dir is None else root_dir
        data_source = ACSDataSource(
            survey_year=year,
            horizon='1-Year',
            survey='person',
            root_dir=data_dir
        )
        acs_data = data_source.get_data(states=state, download=True)
        features = ACSMobility.features
        target = ACSMobility.target
        categorical_columns = ['MAR','SEX','DIS','ESP','CIT','MIL','ANC','NATIVITY','RELP','DEAR','DEYE','DREM','RAC1P','GCL','COW','ESR']
        numerical_columns = ['AGEP', 'SCHL', 'PINCP', 'WKHP', 'JWMNP']

        if with_nulls:
            X_data = acs_data[features]
        else:
            X_data = acs_data[features].apply(lambda x: np.nan_to_num(x, -1))

        filtered_X_data = X_data[categorical_columns].astype('str')
        for col in numerical_columns:
            filtered_X_data[col] = X_data[col]

        y_data = acs_data[target].apply(lambda x: int(x == 1))

        columns_with_nulls = filtered_X_data.columns[filtered_X_data.isna().any().to_list()].to_list()

        super().__init__(
            full_df=acs_data,
            target=target,
            numerical_columns=numerical_columns,
            categorical_columns=categorical_columns,
            X_data=filtered_X_data,
            y_data=y_data,
            columns_with_nulls=columns_with_nulls,
        )

    def update_X_data(self, X_data):
        """
        To save simulated nulls
        """
        self.X_data = X_data
        self.columns_with_nulls = self.X_data.columns[self.X_data.isna().any().to_list()].to_list()


class ACSPublicCoverageDataset(BaseDataLoader):
    """
    Dataset class for the public coverage task from the folktables dataset.
    Target: binary classification, predict whether a low-income individual, not eligible for Medicare,
        has coverage from public health insurance.
    Source of the dataset: https://github.com/socialfoundations/folktables

    Parameters
    ----------
    state
        State in the US for which to get the data. All states in the US are available.
    year
        Year for which to get the data. Five different years of data collection are available: 2014–2018 inclusive.
    root_dir
        Path to the root directory where to store the extracted dataset or where it is stored.
    with_nulls
        Whether to keep nulls in the dataset or replace them on the new categorical class. Default: False.
    with_filter
        Whether to use a folktables filter for this task. Default: True.
    optimize
        Whether to optimize the dataset size by downcasting categorical columns. Default: True.
    subsample_size
        Subsample size to create based on the input dataset.
    subsample_seed
        Seed for sampling using the sample() method from pandas.

    """
    def __init__(self, state, year, root_dir=None, with_nulls=False, with_filter=True,
                 optimize=True, subsample_size: int = None, subsample_seed: int = None):
        data_dir = pathlib.Path(__file__).parent if root_dir is None else root_dir
        data_source = ACSDataSource(
            survey_year=year,
            horizon='1-Year',
            survey='person',
            root_dir=data_dir
        )
        acs_data = data_source.get_data(states=state, download=True)
        if with_filter:
            acs_data = public_coverage_filter(acs_data)
        if subsample_size:
            acs_data = acs_data.sample(subsample_size, random_state=subsample_seed) if subsample_seed is not None \
                else acs_data.sample(subsample_size)
            acs_data = acs_data.reset_index(drop=True)

        features = ACSPublicCoverage.features
        target = ACSPublicCoverage.target
        categorical_columns = ['SCHL','MAR','SEX','DIS','ESP','CIT','MIG','MIL','ANC','NATIVITY','DEAR','DEYE','DREM','ESR','ST','FER','RAC1P']
        numerical_columns = ['AGEP', 'PINCP']

        if with_nulls is True:
            X_data = acs_data[features]
        else:
            X_data = acs_data[features].apply(lambda x: np.nan_to_num(x, -1))

        if optimize:
            X_data = optimize_data_loading(X_data, categorical_columns)

        optimized_X_data = X_data[categorical_columns].astype('str')
        for col in numerical_columns:
            optimized_X_data[col] = X_data[col]
        y_data = acs_data[target].apply(lambda x: int(x == 1))

        columns_with_nulls = optimized_X_data.columns[optimized_X_data.isna().any().to_list()].to_list()

        super().__init__(
            full_df=optimized_X_data,
            target=target,
            numerical_columns=numerical_columns,
            categorical_columns=categorical_columns,
            X_data=optimized_X_data,
            y_data=y_data,
            columns_with_nulls=columns_with_nulls,
        )

    def update_X_data(self, X_data):
        """
        To save simulated nulls
        """
        self.X_data = X_data
        self.columns_with_nulls = self.X_data.columns[self.X_data.isna().any().to_list()].to_list()


class ACSTravelTimeDataset(BaseDataLoader):
    """
    Dataset class for the travel time task from the folktables dataset.
    Target: binary classification, predict whether a working adult has a travel time to work of greater than 20 minutes.
    Source of the dataset: https://github.com/socialfoundations/folktables

    Parameters
    ----------
    state
        State in the US for which to get the data. All states in the US are available.
    year
        Year for which to get the data. Five different years of data collection are available: 2014–2018 inclusive.
    root_dir
        Path to the root directory where to store the extracted dataset or where it is stored.
    with_nulls
        Whether to keep nulls in the dataset or replace them on the new categorical class. Default: False.

    """
    def __init__(self, state, year, root_dir=None, with_nulls=False):
        data_dir = pathlib.Path(__file__).parent if root_dir is None else root_dir
        data_source = ACSDataSource(
            survey_year=year,
            horizon='1-Year',
            survey='person',
            root_dir=data_dir
        )
        acs_data = data_source.get_data(states=state, download=True)
        features = ACSTravelTime.features
        target = ACSTravelTime.target
        categorical_columns = ['MAR','SEX','DIS','ESP','MIG','RELP','RAC1P','PUMA','ST','CIT','OCCP','POWPUMA','POVPIP']
        numerical_columns = ['AGEP', 'SCHL']

        if with_nulls:
            X_data = acs_data[features]
        else:
            X_data = acs_data[features].apply(lambda x: np.nan_to_num(x, -1))

        filtered_X_data = X_data[categorical_columns].astype('str')
        for col in numerical_columns:
            filtered_X_data[col] = X_data[col]
            
        y_data = acs_data[target].apply(lambda x: int(x > 20))
        columns_with_nulls = filtered_X_data.columns[filtered_X_data.isna().any().to_list()].to_list()

        super().__init__(
            full_df=acs_data,
            target=target,
            numerical_columns=numerical_columns,
            categorical_columns=categorical_columns,
            X_data=filtered_X_data,
            y_data=y_data,
            columns_with_nulls=columns_with_nulls,
        )

    def update_X_data(self, X_data):
        """
        To save simulated nulls
        """
        self.X_data = X_data
        self.columns_with_nulls = self.X_data.columns[self.X_data.isna().any().to_list()].to_list()


class ACSDataset_from_demodq(BaseDataLoader):
    """ Following https://github.com/schelterlabs/demographic-data-quality """
    def __init__(self, state, year, root_dir=None, with_nulls=False, optimize=True):
        """
        Loading task data: instead of using the task wrapper, we subsample the acs_data dataframe on the task features
        We do this to retain the nulls as task wrappers handle nulls by imputing as a special category
        Alternatively, we could have altered the configuration from here:
        https://github.com/zykls/folktables/blob/main/folktables/acs.py
        """
        data_dir = pathlib.Path(__file__).parent if root_dir is None else root_dir
        data_source = ACSDataSource(
            survey_year=year,
            horizon='1-Year',
            survey='person',
            root_dir=data_dir
        )
        acs_data = data_source.get_data(states=state, download=True)
        features =  ['AGEP', 'COW', 'SCHL', 'MAR', 'OCCP', 'POBP', 'RELP', 'WKHP', 'SEX', 'RAC1P']
        target = 'PINCP'
        categorical_columns = ['COW', 'SCHL', 'MAR', 'OCCP', 'POBP', 'RELP', 'SEX', 'RAC1P']
        numerical_columns = ['AGEP', 'WKHP']

        if with_nulls:
            X_data = acs_data[features]
        else:
            X_data = acs_data[features].apply(lambda x: np.nan_to_num(x, -1))

        if optimize:
            X_data = optimize_data_loading(X_data, categorical_columns)

        filtered_X_data = X_data
        y_data = acs_data[target].apply(lambda x: x >= 50000).astype(int)
        columns_with_nulls = filtered_X_data.columns[filtered_X_data.isna().any().to_list()].to_list()

        super().__init__(
            full_df=acs_data,
            target=target,
            numerical_columns=numerical_columns,
            categorical_columns=categorical_columns,
            X_data=filtered_X_data,
            y_data=y_data,
            columns_with_nulls=columns_with_nulls,
        )

    def update_X_data(self, X_data):
        """
        To save simulated nulls
        """
        self.X_data = X_data
        self.columns_with_nulls = self.X_data.columns[self.X_data.isna().any().to_list()].to_list()


def optimize_data_loading(data, categorical):
    """
    Optimizing the dataset size by downcasting categorical columns
    """
    for column in categorical:
        data[column] = pd.to_numeric(data[column], downcast='integer')
    return data



import numpy as np

from enum import Enum
from dataclasses import dataclass


@dataclass
class CountPredictionStatsResponse:
    jitter: float
    means_lst: list
    stds_lst: list
    iqr_lst: list
    mean_ensemble_entropy_lst: list
    overall_entropy_lst: np.ndarray
    statistical_bias_lst: np.ndarray
    per_sample_accuracy_lst: list
    label_stability_lst: list


class ModelSetting(Enum):
    INCREMENTAL = "incremental"
    BATCH = "batch"


class ComputationMode(Enum):
    ERROR_ANALYSIS = "error_analysis"


class ReportType(Enum):
    MULTIPLE_RUNS_MULTIPLE_MODELS = "multiple_runs_multiple_models"
    ONE_RUN_MULTIPLE_MODELS = "one_run_multiple_models"
    ONE_RUN_ONE_MODEL = "one_run_one_model"


INTERSECTION_SIGN = '&'
MODELS_TUNING_SEED = 42
MODELS_TUNING_TEST_SET_FRACTION = 0.2



import random
import typing
import datetime as dt

from io import StringIO
from river import base
from river.stream.iter_csv import DictReader


def ddict2dict(d):
    for k, v in d.items():
        if isinstance(v, dict):
            d[k] = ddict2dict(v)
    return dict(d)


def df_to_stream_buffer(df):
    buffer = StringIO()  # creating an empty buffer
    df.to_csv(buffer, index=False)  # filling that buffer
    buffer.seek(0) # set to the start of the stream

    return buffer


def iter_pd_dataset(
        pd_dataset,
        target: typing.Union[str, typing.List[str]] = None,
        converters: dict = None,
        parse_dates: dict = None,
        drop: typing.List[str] = None,
        drop_nones=False,
        fraction=1.0,
        seed: int = None,
        **kwargs,
) -> base.typing.Stream:

    buffer = df_to_stream_buffer(pd_dataset)
    for x in DictReader(fraction=fraction, rng=random.Random(seed), f=buffer, **kwargs):
        if drop:
            for i in drop:
                del x[i]

        # Cast the values to the given types
        if converters is not None:
            for i, t in converters.items():
                if str(t) == "<class 'int'>":
                    # Fix an issue with converting '1.0' to an int type
                    x[i] = int(float(x[i]))
                else:
                    x[i] = t(x[i])

        # Drop Nones
        if drop_nones:
            for i in list(x):
                if x[i] is None:
                    del x[i]

        # Parse the dates
        if parse_dates is not None:
            for i, fmt in parse_dates.items():
                x[i] = dt.datetime.strptime(x[i], fmt)

        # Separate the target from the features
        y = None
        if isinstance(target, list):
            y = {name: x.pop(name) for name in target}
        elif target is not None:
            y = x.pop(target)

        yield x, y



import numpy as np
from munch import DefaultMunch
from sklearn.model_selection import train_test_split

from tests import config_params, compas_dataset_class, compas_without_sensitive_attrs_dataset_class
from virny.utils.common_helpers import validate_config, confusion_matrix_metrics


def test_validate_config_true1(config_params):
    actual = validate_config(config_params)
    assert actual == True


def test_validate_config_true2():
    config_dct = {
        "dataset_name": 'COMPAS',
        "bootstrap_fraction": 0.8,
        "n_estimators": 100,
        "sensitive_attributes_dct": {'sex': 0, 'race': 'Caucasian'},
    }
    config = DefaultMunch.fromDict(config_dct)

    actual = validate_config(config)
    assert actual == True


def test_validate_config_false1():
    config_dct = {
        "dataset_name": 'COMPAS',
        "bootstrap_fraction": 0.8,
        "n_estimators": 100,
        "sensitive_attributes_dct": {'sex': 0, 'race': 'Caucasian', 'sex&race&age': None},
    }
    config = DefaultMunch.fromDict(config_dct)

    try:
        actual = validate_config(config)
    except ValueError:
        actual = False

    assert actual == False


def test_validate_config_false2():
    config_dct = {
        "dataset_name": 'COMPAS',
        "bootstrap_fraction": 1.8,
        "n_estimators": 100,
        "sensitive_attributes_dct": {'sex': 0, 'race': 'Caucasian'},
    }
    config = DefaultMunch.fromDict(config_dct)

    try:
        actual = validate_config(config)
    except ValueError:
        actual = False

    assert actual == False


def test_validate_config_false3():
    config_dct = {
        "dataset_name": 'COMPAS',
        "bootstrap_fraction": 1.8,
        "n_estimators": 100,
        "sensitive_attributes_dct": {'sex': 0, 'sex&race': None},
    }
    config = DefaultMunch.fromDict(config_dct)

    try:
        actual = validate_config(config)
    except ValueError:
        actual = False

    assert actual == False


def test_confusion_matrix_metrics():
    y_true = np.array([0, 1, 0, 1, 0, 1, 0, 1, 0, 1])
    y_preds = np.array([0, 1, 0, 1, 0, 1, 0, 1, 0, 1])

    actual_metrics = confusion_matrix_metrics(y_true, y_preds)

    required_fields = ['TPR', 'TNR', 'PPV', 'FNR', 'FPR', 'Accuracy', 'F1', 'Selection-Rate', 'Positive-Rate']
    for field in required_fields:
        assert field in actual_metrics.keys()



import os
import yaml
import pandas as pd
from munch import DefaultMunch

from virny.utils.common_helpers import validate_config
from virny.configs.constants import INTERSECTION_SIGN


def create_config_obj(config_yaml_path: str):
    """
    Return a config object created based on a config yaml file.

    Parameters
    ----------
    config_yaml_path
        Path to a config yaml file

    """
    with open(config_yaml_path) as f:
        config_dct = yaml.load(f, Loader=yaml.FullLoader)

    config_obj = DefaultMunch.fromDict(config_dct)
    validate_config(config_obj)

    # Fix formatting
    sensitive_attributes_dct_keys = list(config_obj.sensitive_attributes_dct.keys())
    for attr in sensitive_attributes_dct_keys:
        if INTERSECTION_SIGN in attr:
            attrs = attr.strip().split(INTERSECTION_SIGN)
            cleaned_attr = INTERSECTION_SIGN.join(attr.strip() for attr in attrs)
            if cleaned_attr != attr:
                config_obj.sensitive_attributes_dct[cleaned_attr] = config_obj.sensitive_attributes_dct.pop(attr)

    return config_obj


def read_model_metric_dfs(metrics_path, model_names):
    # Read models metrics dfs
    metrics_filenames = [filename for filename in os.listdir(metrics_path)]
    models_metrics_dct = dict()
    for model_name in model_names:
        for filename in metrics_filenames:
            if model_name in filename:
                models_metrics_dct[model_name] = pd.read_csv(f'{metrics_path}/{filename}')
                break

    return models_metrics_dct


def create_models_metrics_dct_from_database_df(model_metric_dfs):
    """
    Create a models_metrics_dct from a metrics dataframe written to a database.
    Return the models_metrics_dct where a key is a model name and a value is a metrics dataframe for the model.

    Parameters
    ----------
    model_metric_dfs
        Path to a config yaml file

    """
    # Create columns based on values in the Subgroup column
    pivoted_model_metric_dfs = model_metric_dfs.pivot(columns='Subgroup', values='Metric_Value',
                                                      index=[col for col in model_metric_dfs.columns
                                                             if col not in ('Subgroup', 'Metric_Value')]).reset_index()
    pivoted_model_metric_dfs = pivoted_model_metric_dfs.rename_axis(None, axis=1)

    # Create a dict of metrics for each model
    models_metrics_dct = dict()
    for model_name in pivoted_model_metric_dfs['Model_Name'].unique():
        models_metrics_dct[model_name] = pivoted_model_metric_dfs[pivoted_model_metric_dfs['Model_Name'] == model_name]

    return models_metrics_dct


def create_models_config_from_tuned_params_df(models_config_for_tuning: dict, tuned_params_df_path: str):
    models_tuned_params_df = pd.read_csv(tuned_params_df_path)
    experiment_models_config = dict()
    for model_name, model_params in models_config_for_tuning.items():
        base_model = create_tuned_base_model(model_params['model'], model_name, models_tuned_params_df)
        experiment_models_config[model_name] = base_model

    return experiment_models_config


def create_tuned_base_model(init_model, model_name, models_tuned_params_df):
    model_params = eval(models_tuned_params_df.loc[models_tuned_params_df['Model_Name'] == model_name,
                                                   'Model_Best_Params'].iloc[0])
    return init_model.set_params(**model_params)



import os
import altair as alt
import numpy as np
import pandas as pd
import datapane as dp
import seaborn as sns
import matplotlib.pyplot as plt
from datetime import datetime, timezone

from virny.configs.constants import ReportType


class MetricsVisualizer:
    """
    Class to create useful visualizations of models metrics.

    Parameters
    ----------
    models_metrics_dct
        Dictionary where keys are model names and values are dataframes of subgroup metrics for each model
    models_composed_metrics_df
        Dataframe of all model composed metrics
    dataset_name
        Name of a dataset that was included in metric filenames and was used for the metrics computation
    model_names
        Metrics for what model names to visualize
    sensitive_attributes_dct
        A dictionary where keys are sensitive attributes names (including attributes intersections),
         and values are privilege values for these attributes

    """
    def __init__(self, models_metrics_dct: dict, models_composed_metrics_df: pd.DataFrame,
                 dataset_name: str, model_names: list, sensitive_attributes_dct: dict):
        sns.set_theme(style="whitegrid")

        self.dataset_name = dataset_name
        self.model_names = model_names
        self.sensitive_attributes_dct = sensitive_attributes_dct
        self.__create_report = False
        self.fairness_metrics_lst = [
            'Accuracy_Parity',
            'Equalized_Odds_TPR',
            'Equalized_Odds_FPR',
            'Disparate_Impact',
            'Statistical_Parity_Difference',
        ]
        self.variance_metrics_lst = [
            'IQR_Parity',
            'Label_Stability_Ratio',
            'Std_Parity',
            'Std_Ratio',
            'Jitter_Parity',
        ]

        # Create models_average_metrics_dct
        models_average_metrics_dct = dict()
        for model_name in model_names:
            columns_to_group = [col for col in models_metrics_dct[model_name].columns
                                if col not in ('Model_Seed', 'Run_Number')]
            models_average_metrics_dct[model_name] = models_metrics_dct[model_name][columns_to_group].groupby(['Metric', 'Model_Name']).mean().reset_index()

        # Create one average metrics df with all model_dfs
        models_average_metrics_df = pd.DataFrame()
        for model_name in models_average_metrics_dct.keys():
            model_average_metrics_df = models_average_metrics_dct[model_name]
            models_average_metrics_df = pd.concat([models_average_metrics_df, model_average_metrics_df])

        models_average_metrics_df = models_average_metrics_df.reset_index(drop=True)

        # Create one metrics df with all model_dfs
        all_models_metrics_df = pd.DataFrame()
        for model_name in models_metrics_dct.keys():
            model_metrics_df = models_metrics_dct[model_name]
            all_models_metrics_df = pd.concat([all_models_metrics_df, model_metrics_df])

        all_models_metrics_df = all_models_metrics_df.reset_index(drop=True)

        self.models_metrics_dct = models_metrics_dct
        self.models_average_metrics_dct = models_average_metrics_dct
        self.all_models_metrics_df = all_models_metrics_df
        self.models_average_metrics_df = models_average_metrics_df
        self.models_composed_metrics_df = models_composed_metrics_df
        self.melted_models_composed_metrics_df = self.models_composed_metrics_df.melt(id_vars=["Metric", "Model_Name"],
                                                                                      var_name="Subgroup",
                                                                                      value_name="Value")
        self.sorted_models_composed_metrics_df = self.melted_models_composed_metrics_df.sort_values(by=['Value'])

    def create_overall_metrics_bar_char(self, metrics_names: list, reversed_metrics_names: list = None,
                                        metrics_title: str = "Overall Metrics"):
        """
        This bar chart includes all defined models and all overall subgroup error and stability metrics,
        which are averaged across multiple runs. Using it, you can compare all models for each subgroup error or stability metric.
        This comparison also includes reversed metrics, in which values closer to zero are better
        since straight and reversed metrics in this plot are converted to the same format -- values closer to one are better.

        Parameters
        ----------
        metrics_names
            List of subgroup metric names to visualize that have a scale from 0 to 1 where closer to 1 is better
        reversed_metrics_names
            List of subgroup metric names to visualize that have a scale from 0 to 1 where closer to 0 is better
        metrics_title
            Title to input metrics (both metrics_names and reversed_metrics_names) to display on the plot

        """
        if reversed_metrics_names is None:
            reversed_metrics_names = []
        metrics_names = set(metrics_names + reversed_metrics_names)

        overall_metrics_df = pd.DataFrame()
        for model_name in self.models_average_metrics_dct.keys():
            model_average_results_df = self.models_average_metrics_dct[model_name].copy(deep=True)
            model_average_results_df = model_average_results_df.loc[model_average_results_df['Metric'].isin(metrics_names)]

            overall_model_metrics_df = pd.DataFrame()
            overall_model_metrics_df['overall'] = model_average_results_df['overall']
            overall_model_metrics_df['metric'] = model_average_results_df['Metric']
            overall_model_metrics_df['model_name'] = model_name
            overall_metrics_df = pd.concat([overall_metrics_df, overall_model_metrics_df])

        overall_metrics_df.loc[overall_metrics_df['metric'].isin(reversed_metrics_names), 'overall'] = \
            1 - overall_metrics_df.loc[overall_metrics_df['metric'].isin(reversed_metrics_names), 'overall']

        models_metrics_chart = (
            alt.Chart(overall_metrics_df).mark_bar().encode(
                alt.Row('metric:N', title=metrics_title),
                alt.Y('model_name:N', axis=None),
                alt.X('overall:Q', axis=alt.Axis(grid=True), title=''),
                alt.Color('model_name:N',
                          scale=alt.Scale(scheme="tableau20"),
                          legend=alt.Legend(title='Model Name',
                                            labelFontSize=13,
                                            titleFontSize=13)
                          )
            )
        ).properties(
            width=500, height=50
        ).configure_headerRow(
            labelAngle=0,
            labelPadding=10,
            labelAlign='left',
            labelFontSize=14,
            titleFontSize=18
        ).configure_axis(
            labelFontSize=14, titleFontSize=18
        )

        return models_metrics_chart

    def create_boxes_and_whiskers_for_models_multiple_runs(self, metrics_lst: list):
        """
        This boxes and whiskers plot is based on overall subgroup error and stability metrics for all defined models
        and results after all runs. Using it, you can see combined information on one plot that includes different models,
         subgroup metrics, and results after multiple runs.

        Parameters
        ----------
        metrics_lst
            List of subgroup metric names to visualize

        """
        to_plot = self.all_models_metrics_df[self.all_models_metrics_df['Metric'].isin(metrics_lst)]

        plt.figure(figsize=(15, 10))
        ax = sns.boxplot(x=to_plot['Metric'],
                         y=to_plot['overall'],
                         hue=to_plot['Model_Name'])

        plt.legend(loc='upper left',
                   ncol=2,
                   fancybox=True,
                   shadow=True)
        plt.xlabel("Metric name")
        plt.ylabel("Metric value")
        fig = ax.get_figure()
        fig.tight_layout()

        if self.__create_report:
            plt.close()
            return fig

    def create_models_metrics_bar_chart(self, metrics_lst: list, metrics_group_name: str, default_plot_metric: str = None):
        if default_plot_metric is None:
            default_plot_metric = metrics_lst[0]

        df_for_model_metrics_chart = self.melted_models_composed_metrics_df.loc[self.melted_models_composed_metrics_df['Metric'].isin(metrics_lst)]
        df_for_model_metrics_chart = df_for_model_metrics_chart.reset_index(drop=True)

        radio_select = alt.selection_single(fields=['Metric'], init={'Metric': default_plot_metric}, empty="none")
        color_condition = alt.condition(radio_select,
                                        alt.Color('Metric:N', legend=None, scale=alt.Scale(scheme="tableau20")),
                                        alt.value('lightgray'))

        models_metrics_chart = (
            alt.Chart(df_for_model_metrics_chart)
            .mark_bar()
            .transform_filter(radio_select)
            .encode(
                x='Value:Q',
                y=alt.Y('Model_Name:N', axis=None),
                color=alt.Color(
                    'Model_Name:N',
                    scale=alt.Scale(scheme="tableau20")
                ),
                row=alt.Row('Subgroup:N', title='Group'),
            )
        )

        select_metric_legend = (
            alt.Chart(df_for_model_metrics_chart)
            .mark_circle(size=200)
            .encode(
                y=alt.Y("Metric:N", axis=alt.Axis(title=f"Select {metrics_group_name} Metric", titleFontSize=15)),
                color=color_condition,
            )
            .add_selection(radio_select)
        )

        color_legend = (
            alt.Chart(df_for_model_metrics_chart)
            .mark_circle(size=200)
            .encode(
                y=alt.Y("Model_Name:N", axis=alt.Axis(title="Model Name", titleFontSize=15)),
                color=alt.Color("Model_Name:N", scale=alt.Scale(scheme="tableau20")),
            )
        )

        return models_metrics_chart, select_metric_legend, color_legend

    def create_fairness_variance_interactive_bar_chart(self):
        """
        This interactive bar chart includes all groups, all composed group fairness and stability metrics,
         and all defined models. Using it, you can select any pair of group fairness and stability metrics and
          compare them across all groups and models. Since this plot is interactive, it saves a lot of space for other plots.
           Also, it could be more convenient to compare each group fairness and stability metric using the interactive mode.
        """
        models_fairness_metrics_chart, select_fairness_metric_legend, fairness_color_legend = \
            self.create_models_metrics_bar_chart(self.fairness_metrics_lst, metrics_group_name="Fairness")

        models_variance_metrics_chart, select_variance_metric_legend, variance_color_legend = \
            self.create_models_metrics_bar_chart(self.variance_metrics_lst, metrics_group_name="Variance")

        return (
                alt.hconcat(
                    alt.vconcat(
                        select_fairness_metric_legend.properties(height=200, width=50),
                        select_variance_metric_legend.properties(height=200, width=50),
                        fairness_color_legend.properties(height=200, width=50),
                    ),
                    models_fairness_metrics_chart.properties(height=200, width=300, title="Fairness Metric Plot"),
                    models_variance_metrics_chart.properties(height=200, width=300, title="Variance Metric Plot"),
                )
        )

    @staticmethod
    def _create_sorted_matrix_by_rank(model_metrics_matrix) -> np.array:
        models_distances_matrix = model_metrics_matrix.copy(deep=True).T
        metric_names = models_distances_matrix.columns
        for metric_name in metric_names:
            if 'impact' in metric_name.lower() or 'ratio' in metric_name.lower():
                models_distances_matrix[metric_name] = models_distances_matrix[metric_name] - 1
            models_distances_matrix[metric_name] = models_distances_matrix[metric_name].abs()

        models_distances_matrix = models_distances_matrix.T
        sorted_matrix_by_rank = np.argsort(np.argsort(models_distances_matrix, axis=1), axis=1)
        return sorted_matrix_by_rank

    def create_model_rank_heatmap(self, model_metrics_matrix, sorted_matrix_by_rank, num_models: int):
        """
        This heatmap includes all group fairness and stability metrics and all defined models.
        Using it, you can visually compare all models across all group metrics. On this plot,
        colors display ranks where 1 is the best model for the metric. These ranks are conditioned
        on difference or ratio operations used to create these group metrics:

        1) if the metric is created based on the difference operation, closer values to zero have ranks that are closer to the first rank

        2) if the metric is created based on the ratio operation, closer values to one have ranks that are closer to the first rank

        Parameters
        ----------
        model_metrics_matrix
            Matrix of model metrics values where indexes are group metric names and columns are model names
        sorted_matrix_by_rank
            Matrix of model ranks per metric where indexes are group metric names and columns are model names
        num_models
            Number of models to visualize

        """
        matrix_width = num_models * 3
        matrix_height = model_metrics_matrix.shape[0] // 3
        plt.figure(figsize=(matrix_width, matrix_height))
        rank_colors = sns.color_palette("coolwarm", n_colors=num_models).as_hex()[::-1]
        ax = sns.heatmap(sorted_matrix_by_rank, annot=model_metrics_matrix, cmap=rank_colors,
                         fmt = '', annot_kws={'color': 'black', 'alpha': 0.7})
        ax.set(xlabel="", ylabel="")
        ax.xaxis.tick_top()

        cbar = ax.collections[0].colorbar
        model_ranks = [idx for idx in range(num_models)]
        cbar.set_ticks([float(idx) for idx in model_ranks])
        tick_labels = [str(idx + 1) for idx in model_ranks]
        tick_labels[0] = tick_labels[0] + ', best'
        tick_labels[-1] = tick_labels[-1] + ', worst'
        cbar.set_ticklabels(tick_labels)
        cbar.set_label('Model Ranks')

        if self.__create_report:
            plt.close()
            return ax

        ax.set_title('Model Ranks Based On Group Fairness and Stability Metrics', fontsize=20)

    def create_total_model_rank_heatmap(self, sorted_matrix_by_rank, num_models):
        """
        This heatmap includes all defined models and sums of their fairness and stability ranks.
        On this plot, colors display sums of ranks for one model. If the sum is smaller,
        the model has better fairness or stability characteristics than other models.
        Using this plot, you can visually compare all models for fairness and stability characteristics.

        Parameters
        ----------
        sorted_matrix_by_rank
            Matrix of model ranks per metric where indexes are group metric names and columns are model names
        num_models
            Number of models to visualize

        """
        total_model_ranks = dict()
        matrix_fairness_metrics = [metric_name for metric_name in sorted_matrix_by_rank[self.model_names[0]].index
                               if metric_name[:metric_name.rfind('_')] in self.fairness_metrics_lst]
        matrix_variance_metrics = [metric_name for metric_name in sorted_matrix_by_rank[self.model_names[0]].index
                                   if metric_name[:metric_name.rfind('_')] in self.variance_metrics_lst]

        for model_name in self.model_names:
            model_ranks = dict()
            model_ranks['Fairness_Ranks_Sum'] = np.sum(sorted_matrix_by_rank[model_name][matrix_fairness_metrics] + 1)
            model_ranks['Variance_Ranks_Sum'] = np.sum(sorted_matrix_by_rank[model_name][matrix_variance_metrics] + 1)
            total_model_ranks[model_name] = model_ranks

        total_model_ranks_df = pd.DataFrame(total_model_ranks).T

        matrix_width = 6
        matrix_height = num_models // 2
        plt.figure(figsize=(matrix_width, matrix_height))
        ax = sns.heatmap(total_model_ranks_df, annot=True, cmap="coolwarm_r", fmt = '')
        ax.set(xlabel="", ylabel="")
        ax.xaxis.tick_top()

        if self.__create_report:
            plt.close()
            return ax

        ax.set_title('Total Ranks Sum For Group Fairness and Stability Metrics', fontsize=15)

    def create_model_rank_heatmaps(self, metrics_lst: list, groups_lst):
        """
        Create model rank and total model rank heatmaps.

        Parameters
        ----------
        metrics_lst
            List of group metric names to visualize
        groups_lst
            List of sensitive attributes

        """
        results = {}
        num_models = len(self.model_names)
        for metric in metrics_lst:
            for group in groups_lst:
                group_metric = metric + '_' + group
                results[group_metric] = dict()
                sorted_model_names_arr = self.sorted_models_composed_metrics_df[
                    (self.sorted_models_composed_metrics_df.Metric == metric) &
                    (self.sorted_models_composed_metrics_df.Subgroup == group)
                    ]['Model_Name'].values
                # Add values to results dict
                for idx, model_name in enumerate(sorted_model_names_arr):
                    metric_value = self.sorted_models_composed_metrics_df[
                        (self.sorted_models_composed_metrics_df.Metric == metric) &
                        (self.sorted_models_composed_metrics_df.Subgroup == group) &
                        (self.sorted_models_composed_metrics_df.Model_Name == model_name)
                        ]['Value'].values[0]
                    metric_value = round(metric_value, 3)
                    results[group_metric][model_name] = metric_value

        model_metrics_matrix = pd.DataFrame(results).T
        sorted_matrix_by_rank = MetricsVisualizer._create_sorted_matrix_by_rank(model_metrics_matrix)
        model_rank_heatmap = self.create_model_rank_heatmap(model_metrics_matrix, sorted_matrix_by_rank, num_models)
        total_model_rank_heatmap = self.create_total_model_rank_heatmap(sorted_matrix_by_rank, num_models)
        if self.__create_report:
            return model_rank_heatmap, total_model_rank_heatmap

    def create_html_report(self, report_type: ReportType, report_save_path: str):
        """
        Create Fairness and Stability Report depending on report type.
        It includes visualizations and helpful details to them.
        """
        # Create a directory if it does not exist
        if not os.path.exists(report_save_path):
            os.makedirs(report_save_path, exist_ok=True)

        self.__create_report = True

        # Create plots
        fairness_overall_metrics_bar_chart = self.create_overall_metrics_bar_char(
            metrics_names=['TPR', 'PPV', 'Accuracy', 'F1', 'Selection-Rate', 'Positive-Rate'],
            metrics_title="Fairness Metrics"
        )
        variance_overall_metrics_bar_chart = self.create_overall_metrics_bar_char(
            metrics_names=['Label_Stability'],
            reversed_metrics_names=['Std', 'IQR', 'Jitter'],
            metrics_title="Stability Metrics"
        )
        interactive_bar_chart = self.create_fairness_variance_interactive_bar_chart()
        model_rank_heatmap, total_model_rank_heatmap = \
            self.create_model_rank_heatmaps(metrics_lst=self.fairness_metrics_lst + self.variance_metrics_lst,
                                            groups_lst=self.sensitive_attributes_dct.keys())

        # Set descriptions for the report
        general_desc = dp.Text(
            f"**Date of creation**: {datetime.now().strftime('%m/%d/%Y, %H:%M:%S')}\n\n\n"
            "This report was created based on the following input arguments:\n"
            f"* __Dataset name__: {self.dataset_name}\n"
            f"* __Model names__: {self.model_names}\n"
            f"* __Sensitive attributes__: {list(self.sensitive_attributes_dct.keys())}\n"
        )
        composed_metrics_desc = dp.Text(
            "Below you can find a dataframe of composed group metrics for all defined models and sensitive attributes.\n"
        )
        boxes_and_whiskers_plot_desc = dp.Text(
            "The below boxes and whiskers plot is based on _overall_ subgroup error and stability metrics for all defined models and results after all runs.\n"
            "This plot can give you the following benefits:\n"
            "* You can see combined information on one plot that includes different models, subgroup metrics, and results after multiple runs\n"
            "* You can see all quartiles for each model metric based on multiple runs\n"
            "* You can compare different models for each metric\n"
            "* You can see the stability of each model metric\n"
        )
        overall_metrics_desc = dp.Text(
            "The below bar chart includes all defined models and all _overall_ subgroup error and stability metrics, which are averaged across multiple runs.\n"
            "This plot can give you the following benefits:\n"
            "* You can compare all models for each subgroup error or stability metric\n"
            "* This comparison also includes reversed metrics, in which values closer to zero are better "
            "since straight and reversed metrics in this plot are converted to the same format -- values closer to one are better\n"
        )
        individual_metrics_interactive_bar_chart_desc = dp.Text(
            "The below interactive bar chart includes all groups, all composed group fairness and stability metrics, "
            "and all defined models.\n"
            "This plot can give you the following benefits:\n"
            "* You can select any pair of group fairness and stability metrics and compare them across all groups and models\n"
            "* Since this plot is interactive, it saves a lot of space for other plots. "
            "Also, it could be more convenient to compare each group fairness and stability metric using the interactive mode\n"
        )
        model_ranked_heatmap_desc = dp.Text(
            "The below heatmap includes all group fairness and stability metrics and all defined models.\n"
            "On this plot, colors display ranks where 1 is the best model for the metric. "
            "These ranks are conditioned on difference or ratio operations used to create these group metrics:\n"
            "* If the metric is created based on the difference operation, **closer values to zero** have ranks that are closer to the first rank\n"
            "* If the metric is created based on the ratio operation, **closer values to one** have ranks that are closer to the first rank\n\n"
            "This plot can give you the following benefits:\n"
            "* You can visually compare all models across all group metrics\n"
            "* You can visually understand where one model is better or worse than other models\n"
            "* You can find the best and worst models for each group metric\n"
        )
        overall_model_ranked_heatmap_desc = dp.Text(
            "The below heatmap includes all defined models and sums of their fairness and stability ranks.\n"
            "On this plot, colors display sums of ranks for one model. If the sum is smaller, the model has better fairness or stability characteristics than other models.\n"
            "This plot can give you the following benefits:\n"
            "* You can visually compare all models for fairness and stability characteristics\n"
            "* You can visually understand where one model is better or worse than other models\n"
            "* You can find the best or most balanced model based on fairness or stability metrics\n"
        )

        report_filename = f'{self.dataset_name}_Metrics_Report_{datetime.now(timezone.utc).strftime("%Y%m%d__%H%M%S")}.html'
        if report_type == ReportType.MULTIPLE_RUNS_MULTIPLE_MODELS:
            boxes_and_whiskers_plot = self.create_boxes_and_whiskers_for_models_multiple_runs(
                metrics_lst=['Std', 'IQR', 'Jitter', 'Label_Stability', 'Accuracy', 'TPR', 'TNR', 'FPR', 'FNR']
            )

            dp.Report("# Fairness and Stability Report",
                      general_desc,

                      "## Model Composed Metrics",
                      composed_metrics_desc,
                      dp.DataTable(self.models_composed_metrics_df),

                      "## Boxes and Whiskers Plot Based On Multiple Models Runs",
                      boxes_and_whiskers_plot_desc,
                      dp.Plot(boxes_and_whiskers_plot),

                      "## Overall Fairness and Stability Model Metrics Comparison",
                      overall_metrics_desc,
                      dp.Plot(fairness_overall_metrics_bar_chart, responsive=False),
                      dp.Plot(variance_overall_metrics_bar_chart, responsive=False),

                      "## Fairness and Stability Interactive Bar Chart",
                      individual_metrics_interactive_bar_chart_desc,
                      dp.Plot(interactive_bar_chart),

                      "## Model Ranks Based On Group Fairness and Stability Metrics",
                      model_ranked_heatmap_desc,
                      dp.Plot(model_rank_heatmap, responsive=False),

                      "## Total Ranks Sum For Group Fairness and Stability Metrics",
                      overall_model_ranked_heatmap_desc,
                      dp.Plot(total_model_rank_heatmap, responsive=False),
                      ).save(path=os.path.join(report_save_path, report_filename))
        else:
            dp.Report("# Fairness and Stability Report",
                      general_desc,

                      "## Model Composed Metrics",
                      composed_metrics_desc,
                      dp.DataTable(self.models_composed_metrics_df),

                      "## Overall Fairness and Stability Model Metrics Comparison",
                      overall_metrics_desc,
                      dp.Plot(fairness_overall_metrics_bar_chart, responsive=False),
                      dp.Plot(variance_overall_metrics_bar_chart, responsive=False),

                      "## Fairness and Stability Interactive Bar Chart",
                      individual_metrics_interactive_bar_chart_desc,
                      dp.Plot(interactive_bar_chart),

                      "## Model Ranks Based On Group Fairness and Stability Metrics",
                      model_ranked_heatmap_desc,
                      dp.Plot(model_rank_heatmap, responsive=False),

                      "## Total Ranks Sum For Group Fairness and Stability Metrics",
                      overall_model_ranked_heatmap_desc,
                      dp.Plot(total_model_rank_heatmap, responsive=False),
                      ).save(path=os.path.join(report_save_path, report_filename))

        self.__create_report = False



import itertools
import numpy as np
import pandas as pd
import scipy as sp


def compute_label_stability(predicted_labels: list):
    """
    Label stability is defined as the absolute difference between the number of times the sample is classified as 0 and 1.
    If the absolute difference is large, the label is more stable.
    If the difference is exactly zero, then it's extremely unstable --- equally likely to be classified as 0 or 1.

    Parameters
    ----------
    predicted_labels

    """
    count_pos = sum(predicted_labels)
    count_neg = len(predicted_labels) - count_pos

    return np.abs(count_pos - count_neg) / len(predicted_labels)


def compute_churn(predicted_labels_1: list, predicted_labels_2: list):
    """
    Pairwise stability metric for two model predictions.

    Parameters
    ----------
    predicted_labels_1

    predicted_labels_2

    """
    return np.sum([int(predicted_labels_1[i] != predicted_labels_2[i])
                   for i in range(len(predicted_labels_1))]) / len(predicted_labels_1)


def compute_jitter(models_prediction_labels):
    """
    Jitter is a stability metric that shows how the base model predictions fluctuate.
    Values closer to 0 -- perfect stability, values closer to 1 -- extremely bad stability.

    Parameters
    ----------
    models_prediction_labels

    """
    n_models = len(models_prediction_labels)
    models_idx_lst = [i for i in range(n_models)]
    churns_sum = 0
    for i, j in itertools.combinations(models_idx_lst, 2):
        churns_sum += compute_churn(models_prediction_labels[i], models_prediction_labels[j])

    return churns_sum / (n_models * (n_models - 1) * 0.5)


def compute_entropy_from_predicted_probability(x):
    """
    Compute entropy from predicted probability

    Parameters
    ----------
    x
        Probability of 0 class

    """
    return sp.stats.entropy([x, 1-x], base=2)


def compute_statistical_bias_from_predict_proba(x, y_true):
    """
    Compute statistical bias from predicted probability

    Parameters
    ----------
    x
        Probability of 0 class
    y_true
        True label

    """
    # If x (main prediction) = 0.4, then expected value = 0 * 0.4 + 1 * (1 - 0.4) = 0.6.
    # For true label = 0, we get bias = abs(0 - 0.6) = 0.6.
    # For true label = 1, we get bias = abs(1 - 0.6) = 0.4.
    expected_val = 0 * x + 1 * (1 - x)
    return abs(y_true - expected_val)


def compute_conf_interval(labels):
    """
    Create 95% confidence interval for population mean weight.

    Parameters
    ----------
    labels

    """
    return sp.stats.norm.interval(alpha=0.95, loc=np.mean(labels), scale=sp.stats.sem(labels))


def compute_std_mean_iqr_metrics(results: pd.DataFrame):
    """
    Compute mean, standard deviation, and interquartile range metrics.

    Parameters
    ----------
    results

    """
    means_lst = results.mean().values
    stds_lst = results.std().values
    iqr_lst = sp.stats.iqr(results, axis=0)

    return means_lst, stds_lst, iqr_lst


def compute_per_sample_accuracy(y_test, results):
    """
    Compute per-sample accuracy for each model predictions.

    Return per_sample_accuracy and label_stability (refer to https://www.osti.gov/servlets/purl/1527311)

    Parameters
    ----------
    y_test
        y test dataset
    results
        `results` variable from count_prediction_stats()

    """
    per_sample_predictions = {}
    label_stability = []
    per_sample_accuracy = []
    acc = None
    for sample in range(len(y_test)):
        per_sample_predictions[sample] =  [int(x<0.5) for x in results[sample].values]
        label_stability.append(compute_label_stability(per_sample_predictions[sample]))

        if y_test[sample] == 1:
            acc = np.mean(per_sample_predictions[sample])
        elif y_test[sample] == 0:
            acc = 1 - np.mean(per_sample_predictions[sample])
        if acc is not None:
            per_sample_accuracy.append(acc)

    return per_sample_accuracy, label_stability



import os
import numpy as np
import pandas as pd

from copy import deepcopy
from tqdm.notebook import tqdm
from abc import ABCMeta, abstractmethod

from virny.custom_classes.custom_logger import get_logger
from virny.utils.data_viz_utils import plot_generic
from virny.utils.stability_utils import generate_bootstrap
from virny.utils.stability_utils import count_prediction_stats, compute_std_mean_iqr_metrics


class AbstractOverallVarianceAnalyzer(metaclass=ABCMeta):
    """
    Abstract class for an analyzer that computes overall variance metrics for subgroups.

    Parameters
    ----------
    base_model
        Base model for stability measuring
    base_model_name
        Model name like 'HoeffdingTreeClassifier' or 'LogisticRegression'
    bootstrap_fraction
        [0-1], fraction from train_pd_dataset for fitting an ensemble of base models
    X_train
        Processed features train set
    y_train
        Targets train set
    X_test
        Processed features test set
    y_test
        Targets test set
    dataset_name
        Name of dataset, used for correct results naming
    n_estimators
        Number of estimators in ensemble to measure base_model stability
    verbose
        [Optional] Level of logs printing. The greater level provides more logs.
         As for now, 0, 1, 2 levels are supported.

    """

    def __init__(self, base_model, base_model_name: str, bootstrap_fraction: float,
                 X_train: pd.DataFrame, y_train: pd.DataFrame, X_test: pd.DataFrame, y_test: pd.DataFrame,
                 dataset_name: str, n_estimators: int, verbose: int = 0):
        self.base_model = base_model
        self.base_model_name = base_model_name
        self.bootstrap_fraction = bootstrap_fraction
        self.dataset_name = dataset_name
        self.n_estimators = n_estimators
        self.models_lst = [deepcopy(base_model) for _ in range(n_estimators)]
        self.models_predictions = None

        self._verbose = verbose
        self.__logger = get_logger(verbose)

        self.X_train = X_train
        self.y_train = y_train
        self.X_test = X_test
        self.y_test = y_test

        # Metrics
        self.mean = None
        self.std = None
        self.iqr = None
        self.aleatoric_uncertainty = None
        self.overall_uncertainty = None
        self.statistical_bias = None
        self.jitter = None
        self.per_sample_accuracy = None
        self.label_stability = None

    @abstractmethod
    def _fit_model(self, classifier, X_train, y_train):
        pass

    @abstractmethod
    def _batch_predict(self, classifier, X_test):
        pass

    @abstractmethod
    def _batch_predict_proba(self, classifier, X_test):
        pass

    def compute_metrics(self, make_plots: bool = False, save_results: bool = True, with_fit: bool = True):
        """
        Measure metrics for the base model. Display plots for analysis if needed. Save results to a .pkl file

        Parameters
        ----------
        make_plots
            bool, if to display plots for analysis
        save_results
            If to save result metrics in a file
        with_fit
            If to fit estimators in bootstrap

        """
        # Quantify uncertainty for the base model
        boostrap_size = int(self.bootstrap_fraction * self.X_train.shape[0])
        self.models_predictions = self.UQ_by_boostrap(boostrap_size, with_replacement=True, with_fit=with_fit)

        # Count metrics based on prediction proba results
        y_preds, uq_labels, prediction_stats = count_prediction_stats(self.y_test.values, self.models_predictions)
        self.__logger.info(f'Successfully computed predict proba metrics')

        self.__update_metrics(means_lst=prediction_stats.means_lst,
                              stds_lst=prediction_stats.stds_lst,
                              iqr_lst=prediction_stats.iqr_lst,
                              mean_ensemble_entropy_lst=prediction_stats.mean_ensemble_entropy_lst,
                              overall_entropy_lst=prediction_stats.overall_entropy_lst,
                              statistical_bias_lst=prediction_stats.statistical_bias_lst,
                              jitter=prediction_stats.jitter,
                              per_sample_accuracy_lst=prediction_stats.per_sample_accuracy_lst,
                              label_stability_lst=prediction_stats.label_stability_lst)

        # Display plots if needed
        if make_plots:
            self.print_metrics()

            # Count metrics based on label predictions to visualize plots
            labels_means_lst, labels_stds_lst, labels_iqr_lst = compute_std_mean_iqr_metrics(uq_labels)
            
            self.__logger.info(f'Successfully computed predict labels metrics')
            per_sample_accuracy_lst = prediction_stats.per_sample_accuracy_lst
            label_stability_lst = prediction_stats.label_stability_lst

            plot_generic(labels_means_lst, labels_stds_lst, "Mean of probability", "Standard deviation", x_lim=1.01,
                         y_lim=0.5, plot_title="Probability mean vs Standard deviation")
            plot_generic(labels_stds_lst, label_stability_lst, "Standard deviation", "Label stability", x_lim=0.5,
                         y_lim=1.01, plot_title="Standard deviation vs Label stability")
            plot_generic(labels_means_lst, label_stability_lst, "Mean", "Label stability", x_lim=1.01, y_lim=1.01,
                         plot_title="Mean vs Label stability")
            plot_generic(per_sample_accuracy_lst, labels_stds_lst, "Accuracy", "Standard deviation", x_lim=1.01,
                         y_lim=0.5, plot_title="Accuracy vs Standard deviation")
            plot_generic(per_sample_accuracy_lst, labels_iqr_lst, "Accuracy", "Inter quantile range", x_lim=1.01,
                         y_lim=1.01, plot_title="Accuracy vs Inter quantile range")

        if save_results:
            self.save_metrics_to_file()
        else:
            return y_preds, self.y_test

    def UQ_by_boostrap(self, boostrap_size: int, with_replacement: bool, with_fit: bool = True) -> dict:
        """
        Quantifying uncertainty of the base model by constructing an ensemble from bootstrapped samples.

        Return a dictionary where keys are models indexes, and values are lists of
         correspondent model predictions for X_test set.

        Parameters
        ----------
        boostrap_size
            Number of records in bootstrap splits
        with_replacement
            Enable replacement or not
        with_fit
            Whether to fit estimators in bootstrap

        """
        models_predictions = {idx: [] for idx in range(self.n_estimators)}
        if self._verbose >= 1:
            print('\n', flush=True)
        self.__logger.info('Start classifiers testing by bootstrap')
        # Remove a progress bar for UQ without estimators fitting
        cycle_range = range(self.n_estimators) if with_fit is False else \
            tqdm(range(self.n_estimators),
                 desc="Classifiers testing by bootstrap",
                 colour="blue",
                 mininterval=10)
        # Train and test each estimator in models_predictions
        for idx in cycle_range:
            classifier = self.models_lst[idx]
            if with_fit:
                X_sample, y_sample = generate_bootstrap(self.X_train, self.y_train, boostrap_size, with_replacement)
                classifier = self._fit_model(classifier, X_sample, y_sample)
            models_predictions[idx] = self._batch_predict_proba(classifier, self.X_test)
            self.models_lst[idx] = classifier

        if self._verbose >= 1:
            print('\n', flush=True)
        self.__logger.info('Successfully tested classifiers by bootstrap')

        return models_predictions

    def __update_metrics(self, means_lst, stds_lst, iqr_lst, mean_ensemble_entropy_lst, overall_entropy_lst,
                         statistical_bias_lst, jitter, per_sample_accuracy_lst, label_stability_lst):
        self.mean = np.mean(means_lst)
        self.std = np.mean(stds_lst)
        self.iqr = np.mean(iqr_lst)
        self.aleatoric_uncertainty = np.mean(mean_ensemble_entropy_lst)
        self.overall_uncertainty = np.mean(overall_entropy_lst)
        self.statistical_bias = np.mean(statistical_bias_lst)
        self.jitter = jitter
        self.per_sample_accuracy = np.mean(per_sample_accuracy_lst)
        self.label_stability = np.mean(label_stability_lst)

    def print_metrics(self):
        precision = 4
        print('\n')
        print("#" * 30, " Stability metrics ", "#" * 30)
        print(f'Mean: {np.round(self.mean, precision)}\n'
              f'Std: {np.round(self.std, precision)}\n'
              f'IQR: {np.round(self.iqr, precision)}\n'
              f'Aleatoric uncertainty: {np.round(self.aleatoric_uncertainty, precision)}\n'
              f'Overall uncertainty: {np.round(self.overall_uncertainty, precision)}\n'
              f'Statistical bias: {np.round(self.statistical_bias, precision)}\n'
              f'Jitter: {np.round(self.jitter, precision)}\n'
              f'Per sample accuracy: {np.round(self.per_sample_accuracy, precision)}\n'
              f'Label stability: {np.round(self.label_stability, precision)}\n\n')

    def get_metrics_dict(self):
        return {
            'Mean': self.mean,
            'Std': self.std,
            'IQR': self.iqr,
            'Aleatoric_Uncertainty': self.aleatoric_uncertainty,
            'Overall_Uncertainty': self.overall_uncertainty,
            'Statistical_Bias': self.statistical_bias,
            'Jitter': self.jitter,
            'Per_Sample_Accuracy': self.per_sample_accuracy,
            'Label_Stability': self.label_stability,
        }

    def save_metrics_to_file(self):
        metrics_to_report = dict()
        metrics_to_report['Dataset_Name'] = [self.dataset_name]
        metrics_to_report['Base_Model_Name'] = [self.base_model_name]
        metrics_to_report['N_Estimators'] = [self.n_estimators]

        metrics_to_report['Mean'] = [self.mean]
        metrics_to_report['Std'] = [self.std]
        metrics_to_report['IQR'] = [self.iqr]
        metrics_to_report['Aleatoric_Uncertainty'] = [self.aleatoric_uncertainty]
        metrics_to_report['Overall_Uncertainty'] = [self.overall_uncertainty]
        metrics_to_report['Statistical_Bias'] = [self.statistical_bias]
        metrics_to_report['Jitter'] = [self.jitter]
        metrics_to_report['Per_Sample_Accuracy'] = [self.per_sample_accuracy]
        metrics_to_report['Label_Stability'] = [self.label_stability]

        metrics_df = pd.DataFrame(metrics_to_report)

        dir_path = os.path.join('..', '..', 'results', 'models_stability_metrics')
        os.makedirs(dir_path, exist_ok=True)

        filename = f"{self.dataset_name}_{self.n_estimators}_estimators_{self.base_model_name}_base_model_stability_metrics.csv"
        metrics_df.to_csv(f'{dir_path}/{filename}', index=False)



import pandas as pd


class BaseFlowDataset:
    """
    Dataset class with custom train and test splits that is used as input for metrics computation interfaces.
    Create your dataset class based on this one to use it for metrics computation interfaces.

    Parameters
    ----------
    init_features_df
        Full train + test non-preprocessed dataset of features without the target column.
         It is used for creating test groups.
    X_train_val
        Train dataframe of features
    X_test
        Test dataframe of features
    y_train_val
        Train dataframe with a target column
    y_test
        Test dataframe with a target column
    target
        Name of the target column name
    numerical_columns
        List of numerical column names
    categorical_columns
        List of categorical column names

    """

    def __init__(self, init_features_df: pd.DataFrame, X_train_val: pd.DataFrame, X_test: pd.DataFrame,
                 y_train_val: pd.DataFrame, y_test: pd.DataFrame,
                 target: str, numerical_columns: list, categorical_columns: list):
        # Validate input sets
        if not isinstance(init_features_df, pd.DataFrame) or not isinstance(X_train_val, pd.DataFrame) \
                or not isinstance(X_test, pd.DataFrame):
            raise ValueError("Input feature sets must be in a pd.DataFrame format")

        self.init_features_df = init_features_df
        self.X_train_val = X_train_val
        self.X_test = X_test
        self.y_train_val = y_train_val
        self.y_test = y_test

        self.numerical_columns = numerical_columns
        self.categorical_columns = categorical_columns
        self.features = numerical_columns + categorical_columns
        self.target = target
        self.columns_with_nulls = self.init_features_df.columns[self.init_features_df.isna().any().to_list()].to_list()



from virny.incremental_ml.river_utils import iter_pd_dataset


class IncrementalPandasDataset:
    """
    Generic data loader that converts a pandas df to a River dataset for incremental models
    """
    def __init__(self, pd_dataset, target, converters: dict):
        self.pd_dataset = pd_dataset
        self.target = target
        self.converters = converters

    def __iter__(self):
        return iter_pd_dataset(
            self.pd_dataset,
            target=self.target,
            converters=self.converters
        )



"""
This module contains functions for computing subgroup variance and error metrics.
"""
from .stability_metrics import (
    compute_std_mean_iqr_metrics,
    compute_churn,
    compute_jitter,
    compute_entropy_from_predicted_probability,
    compute_conf_interval,
    compute_std_mean_iqr_metrics,
    compute_per_sample_accuracy,
)


__all__ = [
    "compute_std_mean_iqr_metrics",
    "compute_churn",
    "compute_jitter",
    "compute_entropy_from_predicted_probability",
    "compute_conf_interval",
    "compute_std_mean_iqr_metrics",
    "compute_per_sample_accuracy",
]



import os
import pytest

from tests import config_params, models_config, ROOT_DIR
from virny.utils.custom_initializers import read_model_metric_dfs
from virny.custom_classes.metrics_composer import MetricsComposer


@pytest.fixture(scope='module')
def models_metrics_dct(models_config):
    metrics_dir_path = os.path.join(ROOT_DIR, 'tests', 'files_for_tests',
                                    'COMPAS_Without_Sensitive_Attributes_Metrics_20230202__094821')
    models_metrics_dct = read_model_metric_dfs(metrics_dir_path, model_names=list(models_config.keys()))
    return models_metrics_dct


# ========================== Test compose_metrics ==========================
def test_compose_metrics_true1(models_metrics_dct, config_params):
    metrics_composer = MetricsComposer(models_metrics_dct, config_params.sensitive_attributes_dct)
    models_composed_metrics_df = metrics_composer.compose_metrics()

    # Check shape
    assert models_composed_metrics_df.shape == (22, 5)

    # Check column names
    assert sorted(models_composed_metrics_df.columns.tolist()) == sorted(['Metric', 'Model_Name', 'sex', 'race', 'sex&race'])

    # Check unique Model_Name
    assert sorted(models_composed_metrics_df['Model_Name'].unique().tolist()) == sorted(['DecisionTreeClassifier', 'LogisticRegression'])

    # Check all metrics presence
    assert sorted(models_composed_metrics_df['Metric'].unique().tolist()) == sorted(['Equalized_Odds_TPR', 'Equalized_Odds_FPR', 'Equalized_Odds_FNR',
                                                                                     'Disparate_Impact', 'Statistical_Parity_Difference', 'Accuracy_Parity',
                                                                                     'Label_Stability_Ratio', 'IQR_Parity', 'Std_Parity',
                                                                                     'Std_Ratio', 'Jitter_Parity'])


def test_compose_metrics_true2(models_metrics_dct, config_params):
    metrics_composer = MetricsComposer(models_metrics_dct, {'sex': 0, 'race': 'Caucasian'})
    models_composed_metrics_df = metrics_composer.compose_metrics()

    # Check shape
    assert models_composed_metrics_df.shape == (22, 4)

    # Check column names
    assert sorted(models_composed_metrics_df.columns.tolist()) == sorted(['Metric', 'Model_Name', 'sex', 'race'])

    # Check unique Model_Name
    assert sorted(models_composed_metrics_df['Model_Name'].unique().tolist()) == sorted(['DecisionTreeClassifier', 'LogisticRegression'])

    # Check all metrics presence
    assert sorted(models_composed_metrics_df['Metric'].unique().tolist()) == sorted(['Equalized_Odds_TPR', 'Equalized_Odds_FPR', 'Equalized_Odds_FNR',
                                                                                     'Disparate_Impact', 'Statistical_Parity_Difference', 'Accuracy_Parity',
                                                                                     'Label_Stability_Ratio', 'IQR_Parity', 'Std_Parity', 'Std_Ratio',
                                                                                     'Jitter_Parity'])



import pandas as pd
from sklearn.compose import ColumnTransformer
from sklearn.preprocessing import StandardScaler
from sklearn.model_selection import train_test_split

from virny.datasets.data_loaders import BaseDataLoader
from virny.custom_classes.base_dataset import BaseFlowDataset


def preprocess_dataset(data_loader: BaseDataLoader, column_transformer: ColumnTransformer,
                       test_set_fraction: float, dataset_split_seed: int) -> BaseFlowDataset:
    """
    Preprocess an input dataset using sklearn ColumnTransformer. Split the dataset on train and test using test_set_fraction.
     Create an instance of BaseFlowDataset.

    Parameters
    ----------
    data_loader
        Instance of BaseDataLoader that contains a target, numerical, and categorical columns.
    column_transformer
        Instance of sklearn ColumnTransformer to preprocess categorical and numerical columns.
    test_set_fraction
        Fraction from 0 to 1. Used to split the input dataset on the train and test sets.
    dataset_split_seed
        Seed for dataset splitting.

    Return
    ----------
    An instance of BaseFlowDataset.

    """
    if test_set_fraction < 0.0 or test_set_fraction > 1.0:
        raise ValueError("test_set_fraction must be a float in the [0.0-1.0] range")

    # Split and preprocess the dataset
    X_train_val, X_test, y_train_val, y_test = train_test_split(data_loader.X_data, data_loader.y_data,
                                                                test_size=test_set_fraction,
                                                                random_state=dataset_split_seed)
    column_transformer = column_transformer.set_output(transform="pandas")  # Set transformer output to a pandas df
    X_train_features = column_transformer.fit_transform(X_train_val)
    X_test_features = column_transformer.transform(X_test)

    return BaseFlowDataset(init_features_df=data_loader.full_df.drop(data_loader.target, axis=1, errors='ignore'),
                           X_train_val=X_train_features,
                           X_test=X_test_features,
                           y_train_val=y_train_val,
                           y_test=y_test,
                           target=data_loader.target,
                           numerical_columns=data_loader.numerical_columns,
                           categorical_columns=data_loader.categorical_columns)


def get_dummies(data: pd.DataFrame, categorical_columns: list, numerical_columns: list):
    """
    Return a dataset made by one-hot encoding for categorical columns and concatenate with numerical columns.

    Parameters
    ----------
    data
        Dataframe for one-hot encoding
    categorical_columns
        List of categorical column names
    numerical_columns
        List of numerical column names

    """
    feature_df = pd.get_dummies(data[categorical_columns], columns=categorical_columns)
    for col in numerical_columns:
        if col in data.columns:
            feature_df[col] = data[col]
    return feature_df


def make_features_dfs(X_train: pd.DataFrame, X_test: pd.DataFrame, dataset: BaseFlowDataset):
    """
    Return preprocessed train and test feature dataframes after one-hot encoding and standard scaling.

    Parameters
    ----------
    X_train

    X_test

    dataset

    """
    X_train_features = get_dummies(X_train, dataset.categorical_columns, dataset.numerical_columns)
    X_test_features = get_dummies(X_test, dataset.categorical_columns, dataset.numerical_columns)

    # Align columns
    features_columns = list(set(X_train_features.columns) & set(X_test_features.columns))
    X_train_features = X_train_features[features_columns]
    X_test_features = X_test_features[features_columns]

    scaler = StandardScaler()
    X_train_features[dataset.numerical_columns] = scaler.fit_transform(X_train_features[dataset.numerical_columns])
    X_test_features[dataset.numerical_columns] = scaler.transform(X_test_features[dataset.numerical_columns])

    return X_train_features, X_test_features



import numpy as np
from sklearn.model_selection import train_test_split

from tests import config_params, compas_dataset_class, compas_without_sensitive_attrs_dataset_class
from virny.preprocessing.basic_preprocessing import make_features_dfs


# ========================== Test make_features_dfs ==========================
def test_make_features_dfs_true1(compas_without_sensitive_attrs_dataset_class, config_params):
    dataset = compas_without_sensitive_attrs_dataset_class
    X_train, X_test, y_train, y_test = train_test_split(dataset.X_data, dataset.y_data,
                                                        test_size=config_params.test_set_fraction,
                                                        random_state=42)
    X_train_features, X_test_features = make_features_dfs(X_train, X_test, dataset)

    # Number of columns and their names are the same in train and test sets
    assert np.array_equal(X_train_features.columns, X_test_features.columns)

    # Check if values changed after normalization
    assert (X_train['juv_fel_count'].values == X_train_features['juv_fel_count'].values).any() == False
    assert (X_test['juv_fel_count'].values == X_test_features['juv_fel_count'].values).any() == False


def test_make_features_dfs_true2(compas_dataset_class, config_params):
    dataset = compas_dataset_class
    X_train, X_test, y_train, y_test = train_test_split(dataset.X_data, dataset.y_data,
                                                        test_size=config_params.test_set_fraction,
                                                        random_state=42)
    X_train_features, X_test_features = make_features_dfs(X_train, X_test, dataset)

    # Number of columns and their names are the same in train and test sets
    assert np.array_equal(X_train_features.columns, X_test_features.columns)

    # Check if values changed after normalization
    assert (X_train['juv_fel_count'].values == X_train_features['juv_fel_count'].values).any() == False
    assert (X_test['juv_fel_count'].values == X_test_features['juv_fel_count'].values).any() == False



import numpy as np
import pandas as pd
import seaborn as sns

from os import listdir
from os.path import isfile, join
from matplotlib import pyplot as plt

from virny.configs.constants import CountPredictionStatsResponse
from virny.utils.data_viz_utils import set_size
from virny.metrics.stability_metrics import compute_std_mean_iqr_metrics, compute_entropy_from_predicted_probability,\
    compute_jitter, compute_per_sample_accuracy, compute_statistical_bias_from_predict_proba


def combine_bootstrap_predictions(bootstrap_predictions: dict, y_test_indexes: np.ndarray):
    """
    Combine predictions generated by estimators in the bootstrap to get final 1D array of predictions.

    Return a pandas series of predictions for each test sample.

    Parameters
    ----------
    bootstrap_predictions
        A dictionary where keys are indexes of bootstrap estimators and values are their predictions for the test set.
    y_test_indexes
        Indexes of the initial test set to keep original row indexes.

    """
    if isinstance(bootstrap_predictions, np.ndarray):
        results = pd.DataFrame(bootstrap_predictions)
    else:
        results = pd.DataFrame(bootstrap_predictions).transpose()

    main_prediction = results.mean().values
    y_preds = np.array([int(x<0.5) for x in main_prediction])

    return pd.Series(y_preds, index=y_test_indexes)


def count_prediction_stats(y_test, uq_results):
    """
    Compute means, stds, iqr, entropy, jitter, label stability, and transform predictions to pd.Dataframe.

    Return a 1D numpy array of predictions, 2D array of each model prediction for y_test, a data structure of metrics.

    Parameters
    ----------
    y_test
        True labels
    uq_results
        2D array of prediction proba for the zero value label by each model

    """
    if isinstance(uq_results, np.ndarray):
        results = pd.DataFrame(uq_results)
    else:
        results = pd.DataFrame(uq_results).transpose()

    means_lst, stds_lst, iqr_lst = compute_std_mean_iqr_metrics(results)
    mean_ensemble_entropy_lst = results.apply(compute_entropy_from_predicted_probability).mean().values

    # Convert predict proba results of each model to correspondent labels.
    # Here we use int(x<0.5) since we use predict_prob()[:, 0] to make predictions.
    # Hence, if a value is, for example, 0.3 --> label == 1, 0.6 -- > label == 0
    uq_labels = results.applymap(lambda x: int(x<0.5))
    jitter = compute_jitter(uq_labels.values)

    main_prediction = results.mean().values
    statistical_bias_lst = np.array(
        [compute_statistical_bias_from_predict_proba(x, y_true) for x, y_true in np.column_stack((main_prediction, y_test))]
    )
    overall_entropy_lst = np.array([compute_entropy_from_predicted_probability(x) for x in main_prediction])

    y_preds = np.array([int(x<0.5) for x in main_prediction])

    per_sample_accuracy_lst, label_stability_lst = compute_per_sample_accuracy(y_test, results)
    prediction_stats = CountPredictionStatsResponse(jitter=jitter,
                                                    means_lst=means_lst,
                                                    stds_lst=stds_lst,
                                                    iqr_lst=iqr_lst,
                                                    mean_ensemble_entropy_lst=mean_ensemble_entropy_lst,
                                                    overall_entropy_lst=overall_entropy_lst,
                                                    statistical_bias_lst=statistical_bias_lst,
                                                    per_sample_accuracy_lst=per_sample_accuracy_lst,
                                                    label_stability_lst=label_stability_lst)

    return y_preds, uq_labels, prediction_stats


def generate_bootstrap(features, labels, boostrap_size, with_replacement=True):
    bootstrap_index = np.random.choice(features.shape[0], size=boostrap_size, replace=with_replacement)
    bootstrap_features = pd.DataFrame(features).iloc[bootstrap_index].values
    bootstrap_labels = pd.DataFrame(labels).iloc[bootstrap_index].values
    if len(bootstrap_features) == boostrap_size:
        return bootstrap_features, bootstrap_labels
    else:
        raise ValueError('Bootstrap samples are not of the size requested')


def display_result_plots(results_dir):
    sns.set_style("darkgrid")
    results = dict()
    filenames = [f for f in listdir(results_dir) if isfile(join(results_dir, f))]

    for filename in filenames:
        results_df = pd.read_csv(results_dir + filename)
        results[f'{results_df.iloc[0]["Base_Model_Name"]}_{results_df.iloc[0]["N_Estimators"]}_estimators'] = results_df

    y_metrics = ['SPD_Race', 'SPD_Sex', 'SPD_Race_Sex', 'EO_Race', 'EO_Sex', 'EO_Race_Sex']
    x_metrics = ['Label_Stability', 'Std']
    for x_metric in x_metrics:
        for y_metric in y_metrics:
            x_lim = 0.3 if x_metric == 'SD' else 1.0
            display_uncertainty_plot(results, x_metric, y_metric, x_lim)


def display_uncertainty_plot(results, x_metric, y_metric, x_lim):
    fig, ax = plt.subplots()
    set_size(15, 8, ax)

    # List of all markers -- https://matplotlib.org/stable/api/markers_api.html
    markers = ['.', 'o', '+', '*', '|', '<', '>', '^', 'v', '1', 's', 'x', 'D', 'P', 'H']
    techniques = results.keys()
    shapes = []
    for idx, technique in enumerate(techniques):
        a = ax.scatter(results[technique][x_metric], results[technique][y_metric], marker=markers[idx], s=100)
        shapes.append(a)

    plt.axhline(y=0.0, color='r', linestyle='-')
    plt.xlabel(x_metric)
    plt.ylabel(y_metric)
    plt.xlim(0, x_lim)
    plt.title(f'{x_metric} [{y_metric}]', fontsize=20)
    ax.legend(shapes, techniques, fontsize=12, title='Markers')

    plt.show()



import os
import pandas as pd

from datetime import datetime, timezone
from abc import ABCMeta, abstractmethod

from virny.configs.constants import ComputationMode


class AbstractSubgroupAnalyzer(metaclass=ABCMeta):
    """
    Abstract class for a subgroup analyzer to compute metrics for subgroups.

    Parameters
    ----------
    X_test
        Processed features test set
    y_test
        Targets test set
    sensitive_attributes_dct
        A dictionary where keys are sensitive attributes names (including attributes intersections),
         and values are privilege values for these attributes
    test_protected_groups
        A dictionary where keys are sensitive attributes, and values input dataset rows
         that are correspondent to these sensitive attributes
    computation_mode
        A mode to compute metrics. It can have two values 'error_analysis' and default (None).

    """
    def __init__(self, X_test: pd.DataFrame, y_test: pd.DataFrame, sensitive_attributes_dct: dict,
                 test_protected_groups: dict, computation_mode: str = None):
        self.sensitive_attributes_dct = sensitive_attributes_dct
        self.X_test = X_test
        self.y_test = y_test
        self.test_protected_groups = test_protected_groups
        self.computation_mode = computation_mode
        self.subgroup_metrics_dict = {}

    @abstractmethod
    def _compute_metrics(self, y_test, y_preds):
        pass

    def _partition_and_compute_metrics(self, y_pred_all, results: dict):
        for group_name in self.test_protected_groups.keys():
            X_test_group = self.test_protected_groups[group_name]
            results[group_name] = self._compute_metrics(self.y_test[X_test_group.index], y_pred_all[X_test_group.index])

        return results

    def _partition_and_compute_metrics_for_error_analysis(self, y_preds, results: dict):
        """
        Partition predictions on correct and incorrect and compute subgroup metrics for each of the partitions.
        Used for the 'error_analysis' mode.

        :param y_preds: a list of predictions
        :param results: a dict to add subgroup metrics for each partition
        """
        for group_name in self.test_protected_groups.keys():
            X_test_group = self.test_protected_groups[group_name]
            group_y_true = self.y_test[X_test_group.index]
            group_y_preds = y_preds[X_test_group.index]

            # Define indexes of each partition of the group: overall group indexes,
            # correct preds group indexes, incorrect preds group indexes
            correct_preds_indexes = group_y_true.index[group_y_true == group_y_preds]
            incorrect_preds_indexes = group_y_true.index[group_y_true != group_y_preds]
            partition_indexes_dct = {
                group_name: X_test_group.index,
                f'{group_name}_correct': correct_preds_indexes,
                f'{group_name}_incorrect': incorrect_preds_indexes,
            }

            # Compute metrics for each group partition
            for group_partition_name, partition_indexes in partition_indexes_dct.items():
                metrics_dct = self._compute_metrics(self.y_test[partition_indexes], y_preds[partition_indexes])
                metrics_dct['Sample_Size'] = len(partition_indexes)
                results[group_partition_name] = metrics_dct

        return results

    def compute_subgroup_metrics(self, y_preds, save_results: bool,
                                 result_filename: str = None, save_dir_path: str = None):
        """
        Compute metrics for each subgroup in self.test_protected_groups using _compute_metrics method.

        Return a dictionary where keys are subgroup names, and values are subgroup metrics.
        
        Parameters
        ----------
        y_preds
            Models predictions
        save_results
            If to save results in a file
        result_filename
            [Optional] Filename for results to save
        save_dir_path
            [Optional] Location where to save the results file

        """
        y_pred_all = pd.Series(y_preds, index=self.y_test.index)

        # Compute overall metrics
        results = dict()
        metrics_dct = self._compute_metrics(self.y_test, y_pred_all)
        metrics_dct['Sample_Size'] = self.y_test.shape[0]
        results['overall'] = metrics_dct

        # Compute metrics for subgroups
        if self.computation_mode == ComputationMode.ERROR_ANALYSIS.value:
            results = self._partition_and_compute_metrics_for_error_analysis(y_pred_all, results)
        else:
            results = self._partition_and_compute_metrics(y_pred_all, results)

        self.subgroup_metrics_dict = results
        if save_results:
            self.save_metrics_to_file(result_filename, save_dir_path)

        return self.subgroup_metrics_dict

    def save_metrics_to_file(self, result_filename: str, save_dir_path: str):
        """
        Parameters
        ----------
        result_filename

        save_dir_path
        """
        metrics_df = pd.DataFrame(self.subgroup_metrics_dict)
        os.makedirs(save_dir_path, exist_ok=True)

        now = datetime.now(timezone.utc)
        date_time_str = now.strftime("%Y%m%d__%H%M%S")
        filename = f"{result_filename}_{date_time_str}.csv"
        metrics_df = metrics_df.reset_index()
        metrics_df.to_csv(f'{save_dir_path}/{filename}', index=False)



"""
This module contains custom classes for metrics computation interfaces.
The purpose is to split metrics computation and visualization pipeline on components
that are highly  customizable for future library features.
"""
from .base_dataset import BaseFlowDataset
from .metrics_composer import MetricsComposer
from .metrics_visualizer import MetricsVisualizer


__all__ = [
    "BaseFlowDataset",
    "MetricsComposer",
    "MetricsVisualizer",
]



import pandas as pd


class MetricsComposer:
    """
    Composer class that combines different subgroup metrics to create group metrics
     such as 'Disparate_Impact' or 'Accuracy_Parity'

    Parameters
    ----------
    models_metrics_dct
        Dictionary where keys are model names and values are dataframes of subgroups metrics for each model
    sensitive_attributes_dct
        A dictionary where keys are sensitive attribute names (including attributes intersections),
         and values are privilege values for these attributes

    """
    def __init__(self, models_metrics_dct: dict, sensitive_attributes_dct: dict):
        self.models_metrics_dct = models_metrics_dct
        self.sensitive_attributes_dct = sensitive_attributes_dct
        self.models_average_metrics_dct = None  # will be created in self.compose_metrics()

    def compose_metrics(self):
        """
        Compose subgroup metrics from self.model_metrics_df.

        Return a dictionary of composed metrics.
        """
        # Create models_average_metrics_dct
        models_average_metrics_dct = dict()
        for model_name in self.models_metrics_dct.keys():
            columns_to_group = [col for col in self.models_metrics_dct[model_name].columns
                                if col not in ('Model_Seed', 'Run_Number')]
            models_average_metrics_dct[model_name] = self.models_metrics_dct[model_name][columns_to_group].groupby(['Metric', 'Model_Name']).mean().reset_index()

        self.models_average_metrics_dct = models_average_metrics_dct

        groups_metrics_dct = dict()
        models_composed_metrics_df = pd.DataFrame()
        for model_name in self.models_average_metrics_dct.keys():
            cfm = self.models_average_metrics_dct[model_name]
            cfm = cfm.set_index('Metric')

            for sensitive_attr in self.sensitive_attributes_dct.keys():
                dis_group = sensitive_attr + '_dis'
                priv_group = sensitive_attr + '_priv'

                groups_metrics_dct[sensitive_attr] = {
                    # Group fairness metrics
                    'Equalized_Odds_TPR': cfm[dis_group]['TPR'] - cfm[priv_group]['TPR'],
                    'Equalized_Odds_FPR': cfm[dis_group]['FPR'] - cfm[priv_group]['FPR'],
                    'Equalized_Odds_FNR': cfm[dis_group]['FNR'] - cfm[priv_group]['FNR'],
                    'Disparate_Impact': cfm[dis_group]['Positive-Rate'] / cfm[priv_group]['Positive-Rate'],
                    'Statistical_Parity_Difference': cfm[dis_group]['Positive-Rate'] - cfm[priv_group]['Positive-Rate'],
                    'Accuracy_Parity': cfm[dis_group]['Accuracy'] - cfm[priv_group]['Accuracy'],
                    # Group stability metrics
                    'Label_Stability_Ratio': cfm[dis_group]['Label_Stability'] / cfm[priv_group]['Label_Stability'],
                    'IQR_Parity': cfm[dis_group]['IQR'] - cfm[priv_group]['IQR'],
                    'Std_Parity': cfm[dis_group]['Std'] - cfm[priv_group]['Std'],
                    'Std_Ratio': cfm[dis_group]['Std'] / cfm[priv_group]['Std'],
                    'Jitter_Parity': cfm[dis_group]['Jitter'] - cfm[priv_group]['Jitter'],
                }

            model_composed_metrics_df = pd.DataFrame(groups_metrics_dct).reset_index()
            model_composed_metrics_df = model_composed_metrics_df.rename(columns={"index": "Metric"})
            model_composed_metrics_df['Model_Name'] = model_name
            models_composed_metrics_df = pd.concat([models_composed_metrics_df, model_composed_metrics_df])

        models_composed_metrics_df = models_composed_metrics_df.reset_index(drop=True)
        return models_composed_metrics_df



import numpy as np

from sklearn.compose import ColumnTransformer
from sklearn.preprocessing import OneHotEncoder
from sklearn.preprocessing import StandardScaler

from tests import config_params, compas_dataset_class, compas_without_sensitive_attrs_dataset_class
from virny.utils.stability_utils import count_prediction_stats, generate_bootstrap
from virny.preprocessing.basic_preprocessing import preprocess_dataset


# ========================== Test count_prediction_stats ==========================
def test_count_prediction_stats_true1():
    y_test = np.array([0, 0, 1, 1, 0, 1, 0, 1, 1, 1])
    uq_results = np.array([[0.6, 0.7, 0.3, 0.4, 0.5, 0.3, 0.7, 0.6, 0.4, 0.4],
                           [0.7, 0.6, 0.4, 0.4, 0.5, 0.3, 0.2, 0.6, 0.4, 0.4]])
    y_preds, uq_labels, prediction_stats = count_prediction_stats(y_test, uq_results)

    mean = np.mean(prediction_stats.means_lst)
    std = np.mean(prediction_stats.stds_lst)
    iqr = np.mean(prediction_stats.iqr_lst)
    aleatoric_uncertainty = np.mean(prediction_stats.mean_ensemble_entropy_lst)
    overall_uncertainty = np.mean(prediction_stats.overall_entropy_lst)
    statistical_bias = np.mean(prediction_stats.statistical_bias_lst)
    per_sample_accuracy = np.mean(prediction_stats.per_sample_accuracy_lst)
    label_stability = np.mean(prediction_stats.label_stability_lst)

    assert np.array_equal(y_preds, np.array([0, 0, 1, 1, 0, 1, 1, 0, 1, 1]))
    assert np.array_equal( uq_labels, np.array([[0, 0, 1, 1, 0, 1, 0, 0, 1, 1], [0, 0, 1, 1, 0, 1, 1, 0, 1, 1]]) )

    alpha = 0.000_001
    assert abs(prediction_stats.jitter - 0.1) < alpha
    assert abs(mean - 0.47000000000000003) < alpha
    assert abs(std - 0.0565685424949238) < alpha
    assert abs(iqr - 0.03999999999999998) < alpha
    assert abs(aleatoric_uncertainty - 0.9345065014636438) < alpha
    assert abs(overall_uncertainty - 0.9560071897163649) < alpha
    assert abs(statistical_bias - 0.42000000000000004) < alpha
    assert abs(per_sample_accuracy - 0.85) < alpha
    assert abs(label_stability - 0.9) < alpha


def test_count_prediction_stats_true2():
    y_test = np.array([0, 0, 1, 1, 0, 1, 0, 1, 1, 1])
    uq_results = np.array([[0.6, 0.7, 0.3, 0.4, 0.5, 0.3, 0.7, 0.6, 0.4, 0.4]])

    try:
        y_preds, uq_labels, prediction_stats = count_prediction_stats(y_test, uq_results)
        actual = True
    except ZeroDivisionError:
        actual = False

    assert actual == False


# ========================== Test generate_bootstrap ==========================
def test_generate_bootstrap_true1(compas_without_sensitive_attrs_dataset_class, config_params):
    column_transformer = ColumnTransformer(transformers=[
        ('categorical_features', OneHotEncoder(handle_unknown='ignore', sparse=False), compas_without_sensitive_attrs_dataset_class.categorical_columns),
        ('numerical_features', StandardScaler(), compas_without_sensitive_attrs_dataset_class.numerical_columns),
    ])
    base_flow_ds = preprocess_dataset(compas_without_sensitive_attrs_dataset_class,
                                      column_transformer,
                                      test_set_fraction=0.2,
                                      dataset_split_seed=42)
    boostrap_size = int(config_params.bootstrap_fraction * base_flow_ds.X_train_val.shape[0])
    X_sample, y_sample = generate_bootstrap(base_flow_ds.X_train_val,
                                            base_flow_ds.y_train_val,
                                            boostrap_size,
                                            with_replacement=True)

    assert X_sample.shape[0] == boostrap_size
    assert X_sample.shape[0] == y_sample.shape[0]
    assert X_sample.shape == (3377, 14)



import numpy as np
import pandas as pd

from virny.analyzers.abstract_overall_variance_analyzer import AbstractOverallVarianceAnalyzer


class BatchOverallVarianceAnalyzer(AbstractOverallVarianceAnalyzer):
    """
    Analyzer to compute subgroup variance metrics for batch learning models.

    Parameters
    ----------
    base_model
        Base model for stability measuring
    base_model_name
        Model name like 'HoeffdingTreeClassifier' or 'LogisticRegression'
    bootstrap_fraction
        [0-1], fraction from train_pd_dataset for fitting an ensemble of base models
    X_train
        Processed features train set
    y_train
        Targets train set
    X_test
        Processed features test set
    y_test
        Targets test set
    target_column
        Name of the target column
    dataset_name
        Name of dataset, used for correct results naming
    n_estimators
        Number of estimators in ensemble to measure base_model stability
    verbose
        [Optional] Level of logs printing. The greater level provides more logs.
         As for now, 0, 1, 2 levels are supported.

    """
    def __init__(self, base_model, base_model_name: str, bootstrap_fraction: float,
                 X_train: pd.DataFrame, y_train: pd.DataFrame, X_test: pd.DataFrame, y_test: pd.DataFrame,
                 target_column: str, dataset_name: str, n_estimators: int, verbose: int = 0):
        super().__init__(base_model=base_model,
                         base_model_name=base_model_name,
                         bootstrap_fraction=bootstrap_fraction,
                         X_train=X_train,
                         y_train=y_train,
                         X_test=X_test,
                         y_test=y_test,
                         dataset_name=dataset_name,
                         n_estimators=n_estimators,
                         verbose=verbose)
        self.target_column = target_column

    def _fit_model(self, classifier, X_train: np.ndarray, y_train: np.ndarray):
        """
        Fit a classifier that is an instance of self.base_model
        """
        return classifier.fit(X_train, y_train)

    def _batch_predict(self, classifier, X_test: pd.DataFrame):
        """
        Predict with the classifier for X_test set and return predictions
        """
        return classifier.predict(X_test)

    def _batch_predict_proba(self, classifier, X_test: pd.DataFrame):
        """
        Predict with the classifier for X_test set and return probabilities for each class for each test point
        """
        return classifier.predict_proba(X_test)[:, 0]



"""API module."""

from . import (
    analyzers,
    custom_classes,
    datasets,
    metrics,
    preprocessing,
    user_interfaces,
    utils,
)

__all__ = [
    "analyzers",
    "custom_classes",
    "datasets",
    "metrics",
    "preprocessing",
    "user_interfaces",
    "utils",
]



import pandas as pd

from virny.configs.constants import INTERSECTION_SIGN


def get_df_condition(df: pd.DataFrame, col: str, dis, include_dis: bool):
    if isinstance(dis, list):
        return df[col].isin(dis) if include_dis else ~df[col].isin(dis)
    else:
        return df[col] == dis if include_dis else df[col] != dis


def partition_by_group_intersectional(df, attrs, dis_values):
    """
    After a partitioning on intersectional groups, a dis group is formed based on the values in sensitive_attributes_dct,
    and a priv group includes all other records, which are not included to a dis group.

    :param df: n initial df
    :param attrs: sensitive attributes
    :param dis_values: disadvantage values for input sensitive attributes

    """
    # Construct complex df conditions
    dis_condition = get_df_condition(df, attrs[0], dis_values[0], include_dis=True)
    for idx in range(1, len(attrs)):
        dis_condition &= get_df_condition(df, attrs[idx], dis_values[idx], include_dis=True)

    priv_condition = (get_df_condition(df, attrs[0], dis_values[0], include_dis=False))
    for idx in range(1, len(attrs)):
        priv_condition |= get_df_condition(df, attrs[idx], dis_values[idx], include_dis=False)

    priv = df[priv_condition]
    dis = df[dis_condition]
    if len(priv) + len(dis) != len(df):
        raise ValueError("Error! Not a partition")

    return priv, dis


def partition_by_group_binary(df, attr, dis_value):
    dis = df[get_df_condition(df, attr, dis_value, include_dis=True)]
    priv = df[get_df_condition(df, attr, dis_value, include_dis=False)]
    if len(priv) + len(dis) != len(df):
        raise ValueError("Error! Not a partition")
    return priv, dis


def check_sensitive_attrs_in_columns(df_columns, sensitive_attributes_dct):
    for sensitive_attr in sensitive_attributes_dct.keys():
        if INTERSECTION_SIGN not in sensitive_attr and sensitive_attr not in df_columns:
            return False
    return True


def create_test_protected_groups(X_test: pd.DataFrame, init_features_df: pd.DataFrame, sensitive_attributes_dct: dict):
    """
    Create protected groups based on a test feature set. Use a disadvantaged group as a reference group.

    Return a dictionary where keys are subgroup names, and values are X_test row indexes correspondent to this subgroup.

    Parameters
    ----------
    X_test
        Test feature set
    init_features_df
        Initial full dataset without preprocessing
    sensitive_attributes_dct
        A dictionary where keys are sensitive attribute names (including attributes intersections),
         and values are disadvantaged values for these attributes

    """
    plain_sensitive_attributes = [attr for attr in sensitive_attributes_dct.keys() if INTERSECTION_SIGN not in attr]
    X_test_with_sensitive_attrs = init_features_df[plain_sensitive_attributes].loc[X_test.index]

    groups = dict()
    for attr in sensitive_attributes_dct.keys():
        attr = attr.strip()
        if INTERSECTION_SIGN in attr:
            single_attrs = attr.split(INTERSECTION_SIGN)
            single_attrs = [single_attr.strip() for single_attr in single_attrs]
            priv_grp_name = INTERSECTION_SIGN.join(single_attrs) + '_priv'
            dis_grp_name = INTERSECTION_SIGN.join(single_attrs) + '_dis'
            groups[priv_grp_name], groups[dis_grp_name] = \
                partition_by_group_intersectional(X_test_with_sensitive_attrs, single_attrs,
                                                  dis_values=[sensitive_attributes_dct[attr] for attr in single_attrs])
        else:
            priv_grp_name = attr + '_priv'
            dis_grp_name = attr + '_dis'
            groups[priv_grp_name], groups[dis_grp_name] = \
                partition_by_group_binary(X_test_with_sensitive_attrs, attr, sensitive_attributes_dct[attr])

        if groups[priv_grp_name].shape[0] == 0:
            raise ValueError(f"Protected group ({priv_grp_name}) from X_test is empty. "
                             f"Please check types of sensitive attributes in config or replace the sensitive attribute")
        if groups[dis_grp_name].shape[0] == 0:
            raise ValueError(f"Protected group ({dis_grp_name}) from X_test is empty. "
                             f"Please check types of sensitive attributes in config or replace the sensitive attribute")

    return groups



import os

from datetime import datetime, timezone
from sklearn.metrics import confusion_matrix
from river import base

from virny.configs.constants import INTERSECTION_SIGN, ModelSetting, ComputationMode


def validate_config(config_obj):
    """
    Validate parameters types and values in config yaml file.

    Extra details:
    * config_obj.model_setting is an optional argument that defines a type of models to use
      to compute fairness and stability metrics. Should be 'batch' or 'incremental'. Default: 'batch'.

    * config_obj.computation_mode is an optional argument that defines a non-default mode for metrics computation.
      Currently, only 'error_analysis' mode is supported.

    Parameters
    ----------
    config_obj
        Object with parameters defined in a yaml file

    """
    # ================== Required parameters ==================
    if not isinstance(config_obj.dataset_name, str):
        raise ValueError('dataset_name must be string')

    if not isinstance(config_obj.bootstrap_fraction, float) \
            or config_obj.bootstrap_fraction < 0.0 \
            or config_obj.bootstrap_fraction > 1.0:
        raise ValueError('bootstrap_fraction must be float in [0.0, 1.0] range')

    if not isinstance(config_obj.n_estimators, int) or config_obj.n_estimators <= 1:
        raise ValueError('n_estimators must be integer greater than 1')

    if not isinstance(config_obj.sensitive_attributes_dct, dict):
        raise ValueError('sensitive_attributes_dct must be python dictionary')

    if isinstance(config_obj.sensitive_attributes_dct, dict):
        intersectional_attrs = [attr for attr in config_obj.sensitive_attributes_dct.keys()
                                if INTERSECTION_SIGN in attr]
        for intersectional_attr in intersectional_attrs:
            intersectional_attr = intersectional_attr.strip()
            attrs = intersectional_attr.split(INTERSECTION_SIGN)
            attrs = [attr.strip() for attr in attrs]
            if len(attrs) != intersectional_attr.count(INTERSECTION_SIGN) + 1:
                raise ValueError(f"Incorrect format for an intersectional attribute name -- {intersectional_attr}."
                                 f"Intersectional signs must be between all attributes in this intersectional attribute.")

            for attr in attrs:
                if attr not in config_obj.sensitive_attributes_dct.keys():
                    raise ValueError('Intersectional attributes in sensitive_attributes_dct must contain '
                                     'single sensitive attributes that also exist in sensitive_attributes_dct')

    # ================== Optional parameters ==================
    if config_obj.model_setting is not None \
            and not isinstance(config_obj.model_setting, str) \
            and config_obj.model_setting not in ModelSetting:
        raise ValueError('model_setting must be a string that is included in the ModelSetting enum. '
                         'Refer to this function documentation for more details!')

    if config_obj.computation_mode is not None \
            and not isinstance(config_obj.computation_mode, str) \
            and config_obj.computation_mode not in ComputationMode:
        raise ValueError('computation_mode must be a string that is included in the ComputationMode enum. '
                         'Refer to this function documentation for more details!')

    return True


def reset_model_seed(model, new_seed, verbose):
    if isinstance(model, base.Classifier): # For incremental models
        model.seed = new_seed
        if verbose >= 1:
            print('Model seed: ', model.seed)
    elif 'random_state' in model.get_params():
        model.set_params(random_state=new_seed)
        if verbose >= 1:
            print('Model seed: ', model.get_params().get('random_state', None))

    return model


def save_metrics_to_file(metrics_df, result_filename, save_dir_path):
    os.makedirs(save_dir_path, exist_ok=True)

    now = datetime.now(timezone.utc)
    date_time_str = now.strftime("%Y%m%d__%H%M%S")
    filename = f"{result_filename}_{date_time_str}.csv"
    metrics_df.to_csv(f'{save_dir_path}/{filename}', index=False)


def confusion_matrix_metrics(y_true, y_preds):
    metrics = {}
    TN, FP, FN, TP = confusion_matrix(y_true, y_preds).ravel()

    metrics['TPR'] = TP/(TP+FN)
    metrics['TNR'] = TN/(TN+FP)
    metrics['PPV'] = TP/(TP+FP)
    metrics['FNR'] = FN/(FN+TP)
    metrics['FPR'] = FP/(FP+TN)
    metrics['Accuracy'] = (TP+TN)/(TP+TN+FP+FN)
    metrics['F1'] = (2*TP)/(2*TP+FP+FN)
    metrics['Selection-Rate'] = (TP+FP)/(TP+FP+TN+FN)
    metrics['Positive-Rate'] = (TP+FP)/(TP+FN)

    return metrics



import pandas as pd

from virny.analyzers.abstract_subgroup_analyzer import AbstractSubgroupAnalyzer
from virny.utils.common_helpers import confusion_matrix_metrics


class SubgroupErrorAnalyzer(AbstractSubgroupAnalyzer):
    """
    Analyzer to compute error metrics for subgroups.

    Parameters
    ----------
    X_test
        Processed features test set
    y_test
        Targets test set
    sensitive_attributes_dct
        A dictionary where keys are sensitive attributes names (including attributes intersections),
         and values are privilege values for these subgroups
    test_protected_groups
        A dictionary where keys are sensitive attributes, and values input dataset rows
         that are correspondent to these sensitive attributes.
    computation_mode
        [Optional] A non-default mode for metrics computation. Should be included in the ComputationMode enum.

    """
    def __init__(self, X_test: pd.DataFrame, y_test: pd.DataFrame,
                 sensitive_attributes_dct: dict, test_protected_groups: dict = None,
                 computation_mode: str = None):
        super().__init__(X_test, y_test, sensitive_attributes_dct, test_protected_groups, computation_mode)

    def _compute_metrics(self, y_test: pd.DataFrame, y_preds: list):
        """
        Compute metrics for subgroups using a confusion matrix
        """
        return confusion_matrix_metrics(y_test, y_preds)



from sklearn.model_selection import train_test_split
from virny.datasets import ACSEmploymentDataset
from virny.utils.protected_groups_partitioning import check_sensitive_attrs_in_columns, create_test_protected_groups

from tests import config_params, folk_emp_config_params, compas_dataset_class, compas_without_sensitive_attrs_dataset_class


def test_check_sensitive_attrs_in_columns_true(config_params):
    df_columns = ['age', 'juv_fel_count', 'juv_misd_count', 'juv_other_count', 'priors_count',
                  'race', 'age_cat_25 - 45', 'age_cat_Greater than 45',
                  'age_cat_Less than 25', 'c_charge_degree_F', 'c_charge_degree_M', 'sex']
    actual = check_sensitive_attrs_in_columns(df_columns,
                                              config_params.sensitive_attributes_dct)
    assert actual == True


def test_check_sensitive_attrs_in_columns_false(config_params):
    df_columns = ['age', 'juv_fel_count', 'juv_misd_count', 'juv_other_count', 'priors_count',
                  'race', 'age_cat_25 - 45', 'age_cat_Greater than 45',
                  'age_cat_Less than 25', 'c_charge_degree_F', 'c_charge_degree_M']
    actual = check_sensitive_attrs_in_columns(df_columns,
                                              config_params.sensitive_attributes_dct)
    assert actual == False


def test_create_test_protected_groups_true(compas_dataset_class, config_params):
    seed = 42
    X_train, X_test, y_train, y_test = train_test_split(compas_dataset_class.X_data,
                                                        compas_dataset_class.y_data,
                                                        test_size=config_params.test_set_fraction,
                                                        random_state=seed)
    actual_test_protected_groups = create_test_protected_groups(X_test, compas_dataset_class.full_df,
                                                                config_params.sensitive_attributes_dct)

    assert len(actual_test_protected_groups) == len(config_params.sensitive_attributes_dct.keys()) * 2

    assert actual_test_protected_groups['sex_priv'].shape[0] == 211
    assert actual_test_protected_groups['sex_dis'].shape[0] == 845
    assert actual_test_protected_groups['race_priv'].shape[0] == 414
    assert actual_test_protected_groups['race_dis'].shape[0] == 642
    assert actual_test_protected_groups['sex&race_priv'].shape[0] == 526
    assert actual_test_protected_groups['sex&race_dis'].shape[0] == 530


def test_create_test_protected_groups_true_without_sensitive_attrs(compas_without_sensitive_attrs_dataset_class, config_params):
    seed = 42
    X_train, X_test, y_train, y_test = train_test_split(compas_without_sensitive_attrs_dataset_class.X_data,
                                                        compas_without_sensitive_attrs_dataset_class.y_data,
                                                        test_size=config_params.test_set_fraction,
                                                        random_state=seed)
    actual_test_protected_groups = create_test_protected_groups(X_test, compas_without_sensitive_attrs_dataset_class.full_df,
                                                                config_params.sensitive_attributes_dct)

    assert len(actual_test_protected_groups) == len(config_params.sensitive_attributes_dct.keys()) * 2

    assert actual_test_protected_groups['sex_priv'].shape[0] == 211
    assert actual_test_protected_groups['sex_dis'].shape[0] == 845
    assert actual_test_protected_groups['race_priv'].shape[0] == 414
    assert actual_test_protected_groups['race_dis'].shape[0] == 642
    assert actual_test_protected_groups['sex&race_priv'].shape[0] == 526
    assert actual_test_protected_groups['sex&race_dis'].shape[0] == 530


def test_create_test_protected_groups_true2(compas_without_sensitive_attrs_dataset_class, config_params):
    new_sensitive_attributes_dct = {'sex': 1, 'race': 'African-American'}

    seed = 42
    X_train, X_test, y_train, y_test = train_test_split(compas_without_sensitive_attrs_dataset_class.X_data,
                                                        compas_without_sensitive_attrs_dataset_class.y_data,
                                                        test_size=config_params.test_set_fraction,
                                                        random_state=seed)
    actual_test_protected_groups = create_test_protected_groups(X_test,
                                                                compas_without_sensitive_attrs_dataset_class.full_df,
                                                                new_sensitive_attributes_dct)

    assert len(actual_test_protected_groups) == len(new_sensitive_attributes_dct.keys()) * 2

    assert actual_test_protected_groups['sex_priv'].shape[0] == 211
    assert actual_test_protected_groups['sex_dis'].shape[0] == 845
    assert actual_test_protected_groups['race_priv'].shape[0] == 414
    assert actual_test_protected_groups['race_dis'].shape[0] == 642



import logging


class CustomHandler(logging.StreamHandler):
    def __init__(self):
        logging.StreamHandler.__init__(self)
        fmt = '%(asctime)s %(filename)-18s %(levelname)-8s: %(message)s'
        fmt_date = '%Y-%m-%d %H:%M:%S'
        formatter = logging.Formatter(fmt, fmt_date)
        self.setFormatter(formatter)


def get_logger(verbose: int = 0):
    logger = logging.getLogger('root')
    logger.setLevel(logging.INFO if verbose >= 1 else logging.WARNING)
    logging.disable(logging.DEBUG)

    if logger.hasHandlers():
        logger.handlers.clear()
    logger.addHandler(CustomHandler())

    return logger
